# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 28 13:37:34 2021

Raami: for operating frame structures

@author: kmela
"""
import os
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
import matplotlib.path as mpath
from datetime import datetime

#import ..framefem.framefem as fem
from metku.framefem import framefem as fem
from metku.raami.frame_node import FrameNode
from metku.raami.frame_member import FrameMember, SteelFrameMember
from metku.raami.frame_loads import PointLoad, LineLoad, PWLineLoad, LoadIDs, LoadCase, LoadCombination
from metku.raami.frame_supports import Support, FixedSupport, XYHingedSupport, XHingedSupport, YHingedSupport
from metku.raami.exports import AbaqusOptions, write_elset, write_nset

#from loadIDs import LoadIDs

# default options for writing to abaqus
# x_monitor: x-coordinate of the fem nodes to follow (or closest)
# n_monitored: number of monitored nodes
def_opts = {'x_monitor':0,'n_monitored':2,'mpc':[],'ecc_elements':[]}

class Raami:
    """ Class for dealing with frame structures.
        The frame can be 2D or 3D, and the frame members
        can be of any material.
        
        The frame is determined by the following data:
            Members
            Nodes
            Supports
            Loads
            Loading combinations
        
        The main operations are:
            1) Structural analysis
            2) Member design
            3) Plotting

        The class is intended to be used with structural optimization too
        
    """
    
    def __init__(self,name="Raami"):
        
        
        self.__name__ = name
        self.dim = 2
        self.members = {}        
        self.nodes = []
        self.member_groups = {}
        self.node_groups = {}
        self.supports = {}
        self.loads = {}
        self.load_ids = []
        self.load_cases = {}
        self.load_combs = {}
    
        self.fem = fem.FrameFEM()
        self.fem_generated = False # Flag for checking if FEM model has been created
        self.is_analysed = False # Flag for checking if structure has been analysed
        
        self.self_weight = False
        
    def __repr__(self):
        
        s = "Frame " + self.__name__ + f" with {len(self.members)} members"
        
        return s
    
    #@property
    #def load_cases(self):
    #    """ Returns a list of load case labels """
    #    return self.load_ids
    
    def weight(self):
        """ Returns weight of the frame in kg """
        #W = sum([mem.weight() for mem in self.members.values()])
        
        W = 0
        for mem in self.members.values():
            W += mem.weight()
        
        return W
    
    def add(self,this):
        """ Adding different things to the frame """
        
        if isinstance(this,FrameNode):
            this.node_id = int(len(self.nodes))            
            self.nodes.append(this)
    
        elif isinstance(this, FrameMember):
            # Give member a unique id
            # id is used in creating elements with wanted cross-sectional properties
            this.mem_id = int(len(self.members))

            # Create a link between the FrameMember object and the frame
            this.frame = self

            # Generate member coordinates
            #this.calc_nodal_coordinates()

            self.members[this.mem_id] = this
        elif isinstance(this,LoadCase):            
            self.load_cases[this.load_id] = this
            if not (this.load_id in self.load_ids):
                self.load_ids.append(this.load_id)
        elif isinstance(this,LoadCombination):
            # Connect the current frame to load combination
            this.frame = self
            self.load_combs[this.comb_id] = this
        # POINTLOADS
        elif isinstance(this, PointLoad):

            """ If a point load with same 'name' is already included
                in the frame, add a number to the end of the name
                of the new load such that each point load has a unique
                name.
                
                NOTE: using 'len' to provide new id number might fail,
                if point loads are deleted such that the number of point loads
                changes along the way.
            """            
            
            if not (this.load_id in self.load_ids):
                # There is no load case for the new load.
                # Add a new load case for this load
                NewLoadCase = LoadCase(load_id=this.load_id,loads = [this])
                self.add(NewLoadCase)
                self.load_ids.append(this.load_id)
            else:
                # If a load case with this.load_id exists, add the new
                # load to that load case
                self.load_cases[this.load_id].add(this)
            
            if this.name in self.loads.keys():
                this.name += str(len(self.loads))
            
            self.loads[this.name] = this

            """ If the location of the point load is in between
                member end nodes, add a node to the corresponding member
            
            for member in self.members.values():
                if member.point_intersection(this.coordinate):
                    member.add_node_coord(this.coordinate)
            """
        # LINELOADS
        elif isinstance(this, LineLoad) or isinstance(this, PWLineLoad):

            if not (this.load_id in self.load_ids):
                
                NewLoadCase = LoadCase(load_id=this.load_id,loads = [this])
                
                self.add(NewLoadCase)
                self.load_ids.append(this.load_id)
            else:
                # If a load case with this.load_id exists, add the new
                # load to that load case
                self.load_cases[this.load_id].add(this)
                
            if this.name in self.loads.keys():
                this.name += str(len(self.loads))
            self.loads[this.name] = this
            
        # SUPPORTS
        elif isinstance(this, Support):
            #print("Adding support")
            # this.supp_id = len(self.supports)
            supp_label = len(self.supports)
            self.supports[supp_label] = this
            # self.supports[this.supp_id] = this
            #self.support_nodes.append(this.coordinate)

            """ If the support is located between end nodes of a member,
                add a node to that member
            
            for member in self.members.values():                
                if member.point_intersection(this.coordinate):                    
                    member.add_node_coord(this.coordinate)
            """
            
    def add_self_weight(self):
        """ Adds self weight to frame's members
        """
        if self.self_weight is False:
            self.self_weight = True
            for member in self.members.values():
                member.add_self_weight()
            # self.add_loads('self_weight')
            # self.generate_frame()
        else:
            self.remove_self_weight()
            self.add_self_weight()    
    
    def remove_self_weight(self):
        """ Removes self weight from frame's members
        """
        self.self_weight = False
        #del (self.line_loads["self_weight"])
        for member in self.members.values():
            member.remove_self_weight()
        
    
    def generate_fem(self):
        """ Generates FEM model 
            The idea is that the FEM model is generated once and subsequently
            the same model is modified.
        """


        if not self.fem_generated:
            self.fem_generated = True


        # Create nodes at the locations of the
        # FrameNodes
        for node in self.nodes:
            if self.dim == 2:
                newNode = self.fem.add_node(node.x,node.y)
            else:
                newNode = self.fem.add_node(node.x,node.y,node.z)

            node.fem_node = newNode

        # Generate elements and internal nodes for members
        for member in self.members.values():
            member.generate_elements(self.fem)

        # Generate supports
        for supp in self.supports.values():
            supp.add_support(self.fem)

        # Generate loads
        for load in self.loads.values():
            load.add_load(self.fem)
            # Creates new loadcase if one with same load_id does not exist
            lcase_ids = [lc.load for lc in self.fem.loadcases.values()]
            if load.load_id not in lcase_ids:
                self.fem.add_loadcase(supp_id=1,load_id=load.load_id)

        # Set nodal degrees of freedom
        self.fem.nodal_dofs()
    
    def clear_fem(self):
        """ Clears all data related to FEM and set up an empty FrameFEM object """        
        self.fem = fem.FrameFEM()
        self.fem_generated = False
        
        for mem in self.members.values():
            mem.clear_fem()
            
    
    def structural_analysis(self, load_id=None, support_method='ZERO', design=True):
        """ Calculates forces and displacements
            
            Parameters
            ----------
            :param load_id: Id of the loads to be added in the calculation model
            
            :type load_id: int / str
        """
        
        # if no load_id is given, use the first load_id available.
        if load_id is None:
            load_id = self.load_ids[0]
        
        # print(f'calculation happening with load id {load_id}')
        if not self.is_analysed:
            self.is_analysed = True
            """ Support ID is always 1! """
            #self.fem.nodal_dofs()

        # If load_id == 'ALL' calculates all load cases
        # calls recursively itself for each case
        if str(load_id).upper() == 'ALL':
            #print("Calculate all cases")
            #lcase_ids = self.load_cases.keys()
            load_ids = []
            for lid in self.load_cases.keys():         
                load_ids.append(lid)
            
            #self.structural_analysis(load_id=load_ids, support_method=support_method)
            self.fem.linear_statics(lcase=load_ids, support_method=support_method)
            
            for lid in load_ids:
                self.calc_nodal_forces(lid)
                
                # Do member design
                if design:
                    self.design_members(lid) 
        else:
            #print('Calculate case:' + str(load_id))            
            self.fem.linear_statics(support_method=support_method,lcase=load_id)
            # Using the FEM results, calculate nodal forces and displacements
            self.calc_nodal_forces(load_id)
            #self.calc_nodal_displacements(load_id)
            # Assign forces to members
            #self.assign_forces(load_id)            
            # Do member design
            if design:
                self.design_members(load_id)            
            # self.alpha_cr, _ = self.f.linear_buckling(k=4)
    
    def calculate(self, load_id=None, support_method='ZERO', design=True):
        """ This method is essentially the same as structural_analysis.
            It is for the purposes of structural optimization, see
            structopt.py -> OptimizationProblem.fea.
            
            OptimizationProblem was written for Frame2D object, which has
            'calculate' as a method for structural optimization.
            
            If OptimizationProblem.fea is changed such that it calls
            'structural_analysis' instead of 'calculate', then this
            method can be deleted.
        """
        
        self.structural_analysis(load_id,support_method,design)
    
    def design_members(self,load_id = LoadIDs['ULS']):
        """ Designs frame members (checks resistance)
        """
        
        for member in self.members.values():            
            member.design(load_id)
    
    def print_member_utilization(self,filename=None,details=False):
        """ Prints utilization ratios for all members """
        
        if filename is None:        
            print(f"** Member utilization ratios for {self.__name__} **\n")
        
            for member in self.members.values():
                member.print_utilization(details=details)
        else:
            with open(filename, 'w') as file:
                file.write(f"** Member utilization ratios for {self.__name__} **\n")
                for member in self.members.values():
                    member.print_utilization(file,details=details)    
                
    
    def optimize_members(self, prof_type="CURRENT", verb=False):
        """ Find minimum smallest profiles for frame members
            for given loads.
        """
                
        kmax = 10
        k = 1
                        
        while k < kmax:            
            explored = []
            MEM_PROFILES_CHANGED = []
            self.structural_analysis('all','REM')
            if verb:
                print(f"Optimize profiles, iteration {k}")
                
            # Go through groups first
            for group in self.member_groups.values():                
                group.optimum_design(prof_type,verb)
                
                for mem in group.members:
                    explored.append(mem)
                
            
            for member in self.members.values():
                #print(member)                
                if not member in explored:
                    if verb:
                        print(member)
                    explored.append(member)
                    MEM_PROFILES_CHANGED.append(member.optimum_design(prof_type,verb))
                                        
                    if not member.symmetry_pair is None:
                        explored.append(member.symmetry_pair)
                    
                    #print(member.cross_section)
            
            # If none of the members changed profile, exit the loop
            if not any(MEM_PROFILES_CHANGED):
                break
                
            k += 1
        
        if verb:
            print(f"Number of iterations: {k}.")
    
    def calc_nodal_forces(self,lcase=LoadIDs['ULS']):
        """ Calculates nodal forces and saves values to
            self.nodal_forces - dict       
        """
        # Iterate through every frame's member's node and calculate forces
        # acting on that node
        for member in self.members.values():
            member.calc_nodal_forces(lcase)
    
    def calc_nodal_displacements(self,lcase=LoadIDs['ULS']):
        """ Calculates nodal displacements and saves values to
            self.nodal_displacements -dict
        """
        # Iterate through every frame's member's node and calculate 
        # displacements for that node
        for member in self.members.values():
            # TODO! IMPLEMENT calc_nodal_displacements for FrameMember
            member.calc_nodal_displacements(self.fem,lcase)
            
    
    def plot(self, print_text=True, show=True,
             loads=True, color=False, axes=None, save=False, mem_dim=False,
             saveopts={'filename':'default.svg','format':'svg','orientation':'landscape', 'papertype':'a3'}):
        """ Plots the frame
            
            Parameters
            ----------
            :param print_text: Set true to print member's profiles and names (default: True)
            :param show: Set true to show the plot (default: True)
            :param loads: Set true to show loads (default: True)
            :param color: Set true to show members' utilization ratio (default: False)

            :type print_text : bool
            :type show: bool
            :type loads: bool
            :type color: bool

            Colors' meaning:

                blue -- member has load
                green -- member can bear its loads
                red -- member breaks under its loads
                black -- member is added, but not designed
        """
        if axes is None:
            if self.dim == 2:
                fig, ax = plt.subplots(1)
            else:
                fig, ax = plt.subplots(1,subplot_kw=dict(projection='3d'))
        else:
            ax = axes


        #if self.is_calculated and color:
        #    color = True

        # Plot members
        for member in self.members.values():
            if self.dim == 2:
                member.plot(print_text, color, ax, mem_dim)
            else:
                member.plot3d(print_text, color, ax, mem_dim)

        # Plot joints
        #for joint in self.joints.values():
        #    joint.plot(color=color)

        # Plot supports
        
        for support in self.supports.values():
            node_coord = support.node.coords
            if support.dofs == [-1, -1, -1] or support.dofs == [1, 1, 1] or support.dofs == [0, 1, 2]:
                marker = 's'
            elif support.dofs == [1]:
                marker = '^'
            elif support.dofs == [0]:
                marker = '>'
            else:
                marker = 'D'
            
            if self.dim == 2:
                ax.scatter(node_coord[0], node_coord[1], s=50, c='k',
                           marker=marker)
            else:
                ax.scatter(node_coord[0], node_coord[1], node_coord[2], zdir='y', s=50, c='k',
                           marker=marker)

        # if self.truss:
        #    self.truss.plot(show=False, print_text=print_text, color=color)
        if loads:
            self.plot_loads()
        
        if self.dim == 2:
            ax.axis('equal')
        else:
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            
        if save:
            #plt.savefig('default.svg', format='svg')
            plt.savefig(saveopts['filename'],format=saveopts['format'],\
                        orientation=saveopts['orientation'],papertype=saveopts['papertype'])
        if show:
            if self.dim == 2:
                plt.axis('equal')
            plt.show()
    
    def plot_nodes(self, print_text=True, show=True, color=False, axes=None, save=False):
        """ Plots the frame
            
            Parameters
            ----------
            :param print_text: Set true to print member's profiles and names (default: True)
            :param show: Set true to show the plot (default: True)
            :param loads: Set true to show loads (default: True)
            :param color: Set true to show members' utilization ratio (default: False)

            :type print_text : bool
            :type show: bool
            :type loads: bool
            :type color: bool

            Colors' meaning:

                blue -- member has load
                green -- member can bear its loads
                red -- member breaks under its loads
                black -- member is added, but not designed
        """
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes


        #if self.is_calculated and color:
        #    color = True

        # Plot members
        for node in self.nodes:            
            node.plot(print_text, 'ok', ax)        
        
        if self.dim == 2:
            ax.axis('equal')
        
        if save:
            plt.savefig('default.svg', format='svg')
        if show:
            if self.dim == 2:
                plt.axis('equal')
            plt.show()
    
    def plot_loads(self):
        """ Plots loads

            Point loads are blue arrows
            Starting point is the given coordinate

            Line loads are red arrows
            Plots loads above given member

            TODO: Load value plotting
        """

        xrange, yrange, _ = self.bounding_box()
        dX = max(xrange)-min(xrange)
        dY = max(yrange)-min(yrange)

        for load in self.loads.values():
            if isinstance(load, PointLoad):
                x, y = load.coordinate                
                scl = max(dX, dY) * 1e-1
                dx, dy, _ = load.v
                plt.arrow(x, y, np.sign(dx) * scl, np.sign(dy) * scl,
                          head_width=scl * 1e-1, ec='b')
            
            elif isinstance(load,LineLoad):
                c0, c1 = load.coordinates
                q1, q2 = load.values
                dq = (q2 - q1) / 10
                x1, y1 = c0
                x2, y2 = c1
                num_arrows = 10
                dx = (x2 - x1) / num_arrows
                dy = (y2 - y1) / num_arrows
                if dx:
                    X = np.arange(x1, x2 + dx, dx)
                else:
                    X = np.ones(11) * x1
                if dy:
                    Y = np.arange(y1, y2 + dy, dy)
                else:
                    Y = np.ones(11) * y1
                scl = max(dX, dY) * 8e-2
                for i, (x, y) in enumerate(zip(X, Y)):
                    q = q1 + dq * i
                    q_scl = q / max(abs(q2), abs(q1))
                    if load.direction == 'y':
                        # Moves arrows above member
                        y += (np.sign(q1) - 2e-1) * scl * q_scl
                        dx = 0
                        dy = -np.sign(q1) * scl * q_scl
                        plt.arrow(x, y, dx, dy,
                                  head_width=scl * 1e-1, ec='r',
                                  head_starts_at_zero=False)
                    else:
                        x -= (np.sign(q1) + 0.25) * scl * q_scl
                        dx = np.sign(q1) * scl * q_scl
                        dy = 0
                        plt.arrow(x, y, dx, dy,
                                  head_width=scl * 1e-1, ec='y')
    def bmd(self, scale=1, save=False, show=True, loads=True, load_id=LoadIDs['ULS']):
        """ Draws bending moment diagram

            Parameters
            ----------
            :param scale: Scaling factor
            :type scale : int
        """
        self.plot(print_text=False, loads=loads, show=False, color=False)
        for member in self.members.values():
            member.bmd(scale, load_id=load_id, show=False, plot_self=False)
        
        if save:
            plt.savefig('bending moment diagram.svg', format='svg')
        if show:
            plt.show()
    
    def xrange(self):
        """ Range of x coordinate values """
        xrange = [node.coords[0] for node in self.nodes]
        
        return (min(xrange),max(xrange))
    
    def yrange(self):
        """ Range of y coordinate values """
        yrange = [node.coords[1] for node in self.nodes]
        
        return (min(yrange),max(yrange))
    
    def bounding_box(self):
        """ Determines the range of coordinate values for different
            axes.
        """
        xrange = [node.coords[0] for node in self.nodes]
        yrange = [node.coords[1] for node in self.nodes]
        
        if self.dim == 3:
            zrange = [node.coords[2] for node in self.nodes]
        else:
            zrange = None
        
        return xrange, yrange, zrange
    
    def symmetry(self,axis='y',sym_value=None):
        """ Detects nodes and members lying symmetrically with respect
            to a given axis.
        """
        
        if axis == 'y':
            if sym_value is None:
                xvalues = self.xrange()
                sym_value = 0.5*(xvalues[0]+xvalues[1])
            
            
            nodes_explored = 0                 
            for i, node in enumerate(self.nodes[:-1]):
                if node.symmetry_node is None:
                    nodes_explored += 1
                    for node2 in self.nodes[i+1:]:
                        if abs(node2.y-node.y) < 1e-6 and abs(abs(node2.x-sym_value)-abs(node.x-sym_value)) < 1e-6:
                            node.symmetry_node = node2
                            node2.symmetry_node = node
                            break
            
            #print(f"Explored nodes: {nodes_explored}.")
                            
        # Find symmetrically lying members:
        # One of the two conditions must be satisfied:
        # a) The both nodes of the members are symmetry pairs, or
        # b) One pair of nodes is symmetric and the other is a shared node
        
        sym_mem_keys = []
        
        for key, mem in self.members.items():            
            if key in sym_mem_keys:
                continue
            else:
                sym_mem_keys.append(key)
                
                if mem.nodes[0].symmetry_node is None and mem.nodes[1].symmetry_node is None:
                    # In this case neither of the member nodes has a symmetry node.
                    # This happens usually with verticals along the symmetry line.
                    continue
                else:
                    # Determine the nodes that the supposed symmetry member should
                    # be linked with.
                    sym_nodes = []
                    for node in mem.nodes:
                        if node.symmetry_node is None:
                            sym_nodes.append(node)
                        else:
                            sym_nodes.append(node.symmetry_node)
                            
                    # Explore the remaining members
                    for sym_key in self.members.keys():
                        if not sym_key in sym_mem_keys:
                            if all([sym_node in self.members[sym_key].nodes for sym_node in sym_nodes]):
                                mem.symmetry_pair = self.members[sym_key]
                                self.members[sym_key].symmetry_pair = mem
                                sym_mem_keys.append(sym_key)
    
    def to_robot(self,filename="Raami"):
        """ Writes the frame as a Autodesk Robot Structural Analysis
            str file.
        """
        
        fname_str = f' File Name:  {filename}  '
        now = datetime.now()
        date_str = now.strftime(" Date: %d/%m/%Y    %H:%M ")
        robot_str = " ROBOT 97 v.34.0     "
        
        top_row = ";+" + '-'*len(fname_str) + '+' + '-'*len(date_str) + "+" + '-'*len(robot_str) + "+"
        mid_row = ";!" + fname_str + "!" + date_str + "!" + robot_str + "!"
        
        # Add point to the end of the filename, if it does not exist yet.
        if filename[1] != '.':
            filename += '.'
        
        with open(filename + 'str', 'w') as f:
            f.write(top_row + "\n")
            f.write(mid_row + "\n")
            f.write(top_row + "\n")
            f.write("ROBOT97 \n\n")
            if self.dim == 3:
                f.write("FRAme SPAce \n\n")
            else:
                f.write("FRAme PLAne \n\n")
                
            f.write("NUMbering DIScontinuous \n\n")
            
            f.write(f'NODes {len(self.nodes)}  ELEments {len(self.members)} \n\n')
            f.write("UNIts \n")
            f.write("LENgth=mm	Force=N \n\n")
            
            f.write("NODes \n")
            for node in self.nodes:
                # It is assumed here that node_id is the ordinal number of the node.
                f.write(f"{node.node_id+1: ^9} {node.x: ^20} {node.y: ^20}\n")              
        
            f.write('\n')
            f.write('ELEments \n')
            f.write(';+------+-------+-------+\n')
            f.write(';! No.  ! STRT  ! END   !\n')
            f.write(';+------+-------+-------+\n')
            
            for mem in self.members.values():
                f.write(f'  {mem.mem_id+1: <6} {mem.nodes[0].node_id+1: ^7} {mem.nodes[1].node_id+1: ^7} \n')

            f.write('\n')
            
            f.write('PROperties \n')
            
            for mem in self.members.values():
                f.write(f' "{mem.material}" \n')
                f.write(f' {mem.mem_id + 1}  ')
                f.write(f' {mem.cross_section.robot()}    \n')

            f.write('\n')
            
            f.write("SUPports \n\n")
            
            for sup in self.supports.values():
                s = f' {sup.node.node_id +1} '
                # sup_dof lists those degrees of freedom that are NOT fixed.
                # 3D cases need to be checked.
                if isinstance(sup,FixedSupport):
                    sup_dof = ' '                    
                elif isinstance(sup,XYHingedSupport):
                    if self.dim == 2:
                        sup_dof = 'RY'
                    else:
                        sup_dof = 'RY, RZ, RX'    
                elif isinstance(sup,XHingedSupport):
                    if self.dim == 2:
                        sup_dof = 'UZ, RY'
                    else:
                        sup_dof = 'UY, UX, RX, RY, RZ'
                elif isinstance(sup,YHingedSupport):
                    if self.dim == 2:
                        sup_dof = 'UX, RY'
                    else:
                        sup_dof = 'UX, UZ, RX, RY, RZ'
                else:
                    raise ValueError(f"Unidentified support type {type(sup)}.")
                
                f.write(s + sup_dof + '\n')
          
            f.write('\n')
            
            f.write("RELeases \n")
            for mem in self.members.values():
                s = mem.robot_releases()
                if not s is None:
                    f.write(s)
            
            f.write('\n')

            
            """
            f.write("LOAds \n")
            
            loads2robot = []
                        
            for i, load_id in enumerate(self.load_ids):
                
                f.write(f"CASe # {i+1} LC1 \n")
                if len(self.line_loads):
                    f.write("ELEments \n")
                    for i in range(num_frames):
                        for lineload in self.line_loads.values():
                            n1, n2 = lineload.member.coordinates
                            if num_frames != 1:
                                n1 = [n1[0], i * s, n1[1]]
                                n2 = [n2[0], i * s, n2[1]]
                            n1_idx = nodes.index(n1) + 1
                            n2_idx = nodes.index(n2) + 1
                            idx = elements.index([n1_idx, n2_idx]) + 1
                            if lineload.direction == 'y':
                                dir = 'PZ'
                            else:
                                dir = 'PX'
                            q0, q1 = lineload.values
                            if q0 != q1:
                                f.write(f' {idx} X=0.0 {dir}={q0} TILl  ')
                                f.write(f'X=1.000  {dir}={q1}      RElative \n')
    
                            else:
                                f.write(f' {idx} {dir}={q0} \n')
                if len(pointloads):
                    f.write("NODes \n")
                    for val in pointloads:
                        f.write(val + "\n")
            
            """
            f.write("END")
                    
    def to_abaqus(self,target_dir='C:/Users/kmela/Data/',filename="Raami",partname='Frame',options=AbaqusOptions()):
        """ Writes the finite element model to ABAQUS 
        
            1. Create folder for the files
            2. Create main file using 'filename'
            3. Write data to main file
            4. For each member, write a file for cross-sectional data        
        """
        
        path = os.path.join(target_dir,filename + '_ABAQUS')
        
        print(path)
        
        try:
            os.mkdir(path)
        except OSError as error:
            print(f"Folder {path} already exists.")
        
        inp_file = path + '/' + filename + '.inp'
        
        secfiles = []
        matfiles = []
        
        # File for linear elastic material models
        elastic_mat_file = path + '/' + filename + '_Materials_Elastic.txt'
        
        # Create material data for elastic material model
        with open(elastic_mat_file,'w') as file:
            file.write(f'**\n** This file contains materials for {filename}\n**\n')
            
            for mem_name, mem in self.members.items():
                if len(options.included_members) > 0 and mem_name in options.included_members:
                    file.write(f'*Material, name={mem_name}_Mat\n')
                    file.write('*Density\n7850.,\n*Elastic\n2.10000e+11, 0.3\n**\n')
            
            for set_name, els in options.elsets.items():
                file.write(f'*Material, name={set_name}_Mat\n')
                file.write('*Density\n7850.,\n*Elastic\n2.10000e+11, 0.3\n**\n')
        
        # File for all section properties
        sections_file = path + '/' + filename + '_Sections.txt'
        
        # Write header for the file that contains cross-section data
        with open(sections_file,'w') as file:
            now = datetime.now()            
            date_str = now.strftime(" Date: %d/%m/%Y    %H:%M ")
            file.write(f'** Cross-sectional data for {filename}\n')
            file.write(f'** Created on {date_str}\n**\n')
            
        for mem_name, mem in self.members.items():
            # Create cross-section a file for each member
            if len(options.included_members) > 0 and mem_name in options.included_members:
                sec_name = mem.cross_section.__repr__().replace('.','_')
                sec_name = sec_name.replace(' ','_')
                sec_name += '_' + mem.material.__repr__()
                secfile = partname + '_' + mem_name + '_' + sec_name + '.txt'
                matfile = partname + '_' + mem_name + '_' + sec_name + '_Mat.txt'
                secfile_path = path + '/' + secfile
                
                #mem.cross_section.abaqus(secfile_path,matname=mem_name,setname=mem_name)
                #secfiles.append(secfile)
                mem.cross_section.abaqus(sections_file,matname=mem_name,setname=mem_name,open_type='a')
                matfiles.append(matfile)
        
        for set_name, els in options.elsets.items():
            # Create cross-section for each element set
            sec_name = els[0].section.__repr__().replace('.','_')
            sec_name = sec_name.replace(' ','_')
            sec_name += '_' + els[0].material.__repr__()
            secfile = partname + '_' + set_name + '_' + sec_name + '.txt'
            matfile = partname + '_' + set_name + '_' + sec_name + '_Mat.txt'
            secfile_path = path + '/' + secfile
            els[0].section.abaqus(sections_file,matname=set_name,setname=set_name,open_type='a')
            #els[0].section.abaqus(secfile_path,matname=set_name,setname=set_name)
            secfiles.append(secfile)
            matfiles.append(matfile)
            
        with open(sections_file,'a') as file:
            file.write('**')
        
        with open(inp_file,'w') as file:
            now = datetime.now()
            print(now)
            date_str = now.strftime(" Date: %d/%m/%Y    %H:%M ")
            file.write(f'** {filename}\n')
            file.write(f'** Created on {date_str}\n')
            file.write('**\n')
            file.write(f'*Part, name={partname}\n')
            # Write nodal coordinates
            file.write('*Node\n')
            for node in self.fem.nodes:
                file.write(f'{node.nid+1:>7g}, {node.x*1e-3:>14.10f},  0.0000000000, {node.y*1e-3:>14.10f}\n')
                    
            # Write elements.
            # Eccentricity elements are handled by multi-point constraints (MPC),
            # so they are excluded here.
            file.write('*Element, type=B31\n')
            for eid, el in enumerate(self.fem.elements):
                if not el in options.ecc_elements:
                    file.write(f'{eid+1:>5g}, {el.nodes[0].nid+1:>6g}, {el.nodes[1].nid+1:>6g}\n')
            
            # Element sets: the element set for each member
            # is for setting the cross-section for the elements
            s1_rel = []
            s2_rel = []
            for mem_name, mem in self.members.items():
                if len(options.included_members) > 0 and mem_name in options.included_members:
                    nel = len(mem.fem_elements)
                    
                    if mem.hinges[0]:
                        # This append does not make sense!
                        s1_rel.append
                    
                    if nel > 1:
                        #file.write(f'*Elset, elset={mem_name}, generate\n')
                        file.write(f'*Elset, elset={mem_name}\n')
                        for i, ele in enumerate(mem.fem_elements):
                            ndx = self.fem.elements.index(ele)+1
                            if i < nel-1:
                                file.write(f'{ndx:>5g}, ')
                            else:
                                file.write(f'{ndx:>5g}')
                            
                            # If member has releases add it to the corresponding
                            # release set. '2' is for rotational release at the first node
                            # and '5' is for rotational release at the second node
                            if 2 in ele.releases:
                                s1_rel.append(ndx)
                            elif 5 in ele.releases:
                                s2_rel.append(ndx)
                                                    
                        file.write('\n')
                    else:
                        file.write(f'*Elset, elset={mem_name}\n')
                        e1 = self.fem.elements.index(mem.fem_elements[0])+1
                        file.write(f'{e1:>5g}\n')
                            #e1 = self.fem.elements.index(mem.fem_elements[0])+1
                            #e2 = self.fem.elements.index(mem.fem_elements[-1])+1
                            #file.write(f'{e1:>5g}, {e2:>5g},   1\n')
            
            # Gap elements (if any)
            top_ndx = []            
            for el in options.top_gap_elements:
                top_ndx.append(self.fem.elements.index(el)+1)
            
            if len(top_ndx) > 1:
                write_elset(file, 'TopChordGaps', top_ndx)
                
            bottom_ndx = []
            for el in options.bottom_gap_elements:
                bottom_ndx.append(self.fem.elements.index(el)+1)
            
            if len(bottom_ndx) > 1:
                write_elset(file, 'BottomChordGaps', bottom_ndx)
                
            # If there are any other element sets, they are printed here
            for set_name, els in options.elsets.items():                
                el_ndx = [self.fem.elements.index(el)+1 for el in els]                
                write_elset(file, set_name, el_ndx)
                
            # Release sets
            file.write('**\n')
            
            write_elset(file,"S1Release",s1_rel)
            file.write('**\n')
            write_elset(file,"S2Release",s2_rel)
           
            # Include cross-sectional data from file 'sections_file'
            file.write(f'*Include, Input={sections_file}\n')
            #*Include, Input=S023C005M002_RedMat.txt            
            #for secfile in secfiles:
            #    file.write(f'*Include, Input={secfile}\n')
            
    
            # State element releases            
            file.write('**\n*RELEASE\n')
            file.write('S1Release, S1, M1\n')
            file.write('S2Release, S2, M1\n')

            #[elementti setin nimi alkupää], S1, M1
            #[elementti setin nimi loppupää], S2, M1
            
            file.write('*End Part\n')
            
            # Start Assembly part of the file. This includes
            # 1) Supports
            # 2) Multi-point constraints (hinges)
            # 3) Loads
            file.write('**\n**\n** ASSEMBLY \n**\n')
            file.write(f'*Assembly, name={partname}_Assembly \n**\n')
            
            # Name of the assembly instance
            insname = filename
            file.write(f'*Instance, name={insname}, part={partname}\n')
            file.write('*End Instance\n**\n')

            # Write element sets for loads
            for set_name, els in options.load_elsets.items():
                if len(els) > 0:
                    el_ndx = [self.fem.elements.index(el)+1 for el in els]                
                    write_elset(file, set_name, el_ndx, insname)
                    
            for set_name, nodes in options.load_nsets.items():
                if len(nodes) > 0:
                    node_ndx = [self.fem.nodes.index(node)+1 for node in nodes]
                    write_nset(file, set_name, node_ndx, insname)

            # Node sets for the following groups
            # 1) Symmetry conditions to prevent displacement in y direction
            #   (all nodes EXCEPT slave nodes of MPC)
            # 2) Supports: a) pinned support, b) slide support, where displacement in z direction
            #   is prevented.
            # 3) Monitored node: node, whose displacement will be traced.
            file.write('** Node set for out-of-plane support\n')
            file.write(f'*Nset, nset=BC_Y_symm_nodes, instance={insname}\n')
            
            # By default, all nodes are master nodes            
            y_symm_nodes = [n.nid+1 for n in self.fem.nodes]
            for mpc in options.mpc:
                # If MPCs have been defined in 'options', the first
                # element in each member of mpc list is the slave node index
                # (in abaqus numbering)
                y_symm_nodes.remove(mpc[0])
            
            i = 0
            while i < len(y_symm_nodes):
                file.write(', '.join(str(r) for r in y_symm_nodes[i:i+16]))
                file.write('\n')
                i += 16
            
            file.write('** Node set for monitored nodes.\n')
            # Find 'n_monitored' closed nodes with respect to the x coordinate in the FEM
            # nodes. These will be monitored
            monitored_nodes_set = 'Monitored_Nodes'
            monitored_nodes = list(np.argsort([abs(n.x-options.x_monitor) for n in self.fem.nodes])[:options.n_monitored])
            file.write(f'*Nset, nset={monitored_nodes_set}, instance={insname}\n')
            file.write(', '.join(str(mn) for mn in monitored_nodes))
            file.write('\n')
            
            # Supported nodes
            supp_fem_nodes = {'fixed':[], 'hinged':[], 'y_supp': [], 'x_supp': []}            
            for supp in self.supports.values():                
                if isinstance(supp,XYHingedSupport):
                    supp_fem_nodes['hinged'].append(supp.node.fem_node.nid)
                elif isinstance(supp,YHingedSupport):
                    supp_fem_nodes['y_supp'].append(supp.node.fem_node.nid)
                elif isinstance(supp,XHingedSupport):
                    supp_fem_nodes['x_supp'].append(supp.node.fem_node.nid)
                elif isinstance(supp,FixedSupport):
                    supp_fem_nodes['fixed'].append(supp.node.fem_node.nid)
            
            #print(supp_fem_nodes)
            
            file.write('** Node sets for supports\n')
            for supp_type, nids in supp_fem_nodes.items():
                nnodes = len(nids)
                if nnodes > 0:
                    file.write(f'*Nset, nset={supp_type}, instance={insname}\n')
                    for i, nid in enumerate(nids):                    
                        if i < nnodes-1:
                            file.write(f'{nid+1:>5g}, ')
                        else:
                            file.write(f'{nid+1:>5g}')
                    file.write('\n')
            
            # Eccentricity elements: write them as MPCs.
            for mpc in options.mpc:
                file.write('*MPC\n')
                file.write(f'BEAM, {insname}.{mpc[0]}, {insname}.{mpc[1]}\n')
            
            file.write('*End Assembly\n')
            
            # Write material files:
            file.write('**\n** MATERIALS\n**\n')
            file.write(f'*Include, Input={elastic_mat_file}\n')
            #for matfile in matfiles:
            #    file.write(f'*Include, Input={matfile}\n')
            
            # Write supports
            file.write('**\n** BOUNDARY CONDITIONS\n**\n')        
            for supp_type, nids in supp_fem_nodes.items():
                nnodes = len(nids)
                if nnodes > 0:
                    file.write('*Boundary\n')
                    if supp_type == 'hinged':
                        supp_code = 'PINNED'
                    elif supp_type == 'fixed':
                        supp_code = 'ENCASTRE'
                    elif supp_type == 'y_supp':
                        supp_code = '3' # pitäisikö olla ZSYMM?
                    elif supp_type == 'x_supp':
                        supp_code = 'XSYMM'
                    file.write(f'{supp_type}, {supp_code}\n')
            
            file.write('** --------------------------------\n')
            file.write('** STEP\n')
            file.write('*Step, name=Buckling, nlgeom=YES, extrapolation=NO, inc=200\n')
            file.write('*Static, riks\n')
            file.write('0.01, 1., 1e-16, 0.05, 1.,\n')

            # Write loads
            file.write('**\n** LOADS\n**\n')
            
            # Self-weight, imposed on all elements
            file.write('*Dload\n')
            file.write(', GRAV, 9.81, 0., 0., -1.\n')
            
            # Write output requests
            file.write('**\n** OUTPUT REQUESTS\n**\n')
            file.write('**FIELD OUTPUT:\n')
            file.write('*Output, field, variable=PRESELECT\n')
            file.write('*Output, field\n')
            file.write('*Node Output\n')
            file.write('U\n')
            file.write('*Element Output, directions=YES\n')
            file.write('E, PEEQ, S, MISES, SF\n')
            
            #file.write('*Output, field\n')
            #file.write(f'*Node Output, nset={monitored_nodes_set}\n')
            #file.write('U1, U2, U3, UR1, UR2, UR3\n')
            
            file.write('**HISTORY OUTPUT:\n')
            file.write('*Output, history, variable=PRESELECT\n')
            file.write('*Output, history\n')
            file.write(f'*Node Output, nset={monitored_nodes_set}\n')
            file.write('U1, U2, U3, UR1, UR2, UR3\n')

            file.write('*End Step\n')
            
if __name__ == '__main__':
    
    from metku.sections.steel.RHS import RHS, SHS
    import cProfile
    
    f = Raami()
    
    L = 5000.0
    H = 2000.0
    nodes = [[0.0,0.0],[0.0,H],[L,H],[L,0.0]]
    
    s = RHS(200,200,6)
    
    for node in nodes:
        f.add(FrameNode(node))
    
    f.add(SteelFrameMember(nodes=f.nodes[:2],section=s,mem_type='column',nel=6))
    f.add(SteelFrameMember(nodes=f.nodes[1:3],section=s,mem_type='beam',nel=6,hinges=[True,True]))
    f.add(SteelFrameMember(nodes=f.nodes[2:4],section=s,mem_type='column',nel=6))
    
    f.add(FixedSupport(f.nodes[0]))
    f.add(FixedSupport(f.nodes[-1]))
    
    #f.add(PointLoad(f.nodes[1], [0, -2000, 0],load_id=LoadIDs.ULS))
    
    f.add(LineLoad(f.members[1],[-20,-20],'y',coord_sys='local',load_id=0,ltype='snow',name='Lumi'))
    f.add(LineLoad(f.members[0],[3,3],'x',coord_sys='global',load_id=1,ltype='wind',name='Tuuli x+'))
    
    f.add(LoadCombination(comb_id=2,comb_type='ULS',load_cases=list(f.load_cases.values())))
    
    f.load_combs[2].combine_loads()
    #f.plot(loads=True)
    f.generate_fem()
    #f.to_robot()
    f.structural_analysis(load_id=2,support_method="REM")
    #f.bmd(scale=10,load_id=2, loads=False)


    f.members[0].bmd(load_id=2)
    #f.optimize_members("CURRENT")
    
            


# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 14:56:32 2022

Optimization variable classes

@author: kmela
"""

import math

from typing import Optional
from metku.optimization.constants import XLB, XUB, INT_TOL, DISC_TOL


class Variable:
    """ Class for optimization variables
    """

    def __init__(self,
                 name: str,
                 lb: float = XLB,
                 ub: float = XUB,
                 target: Optional[dict[str, object | list]] = None,
                 target_property: Optional[str] = None,
                 target_properties: Optional[list[str]] = None,
                 target_object: Optional[object] = None,
                 target_objects: Optional[list[object]] = None,
                 profiles: Optional[list] = None,
                 value: Optional[float] = None,
                 target_fun: Optional[callable] = None):
        """
        Parameters:
            -----------
                :param name: string stating the name of the variable
                :param value: current value of the variable
                :param lb: lower bound
                :param ub: upper bound
                :param target: dict that provides information about the part of
                the structure that the variable is changing
                :param profiles: list of available profiles (optional)

                target = {"property": string, "objects": list}
                    property .. unique identifier that shows which property of
                                the structure is affected by the variable
                                
                                this can also be a list of strings
                    objects .. list of objects (e.g. frame members) that are
                                affected by the variable

                if target is 'None', then the variable does not change anything
                in the structure
                :param target_fun: (optional) dict that provides functions for evaluating a functional relationship
                                    between the variable and its target. For cases, where the variable substitution
                                    is not a direct attribute insertion. For example, when the cross-sectional area
                                    of a member is used as a variable, and it also affects the second moment of area
                                    by some function I = I(A)

                Allowed values for 'property':
                    'AREA' .. cross-sectional area of a member
                    'IY' .. second moment of area of a member (major axis)
                    'IZ' .. second moment of area of a member (minor axis)
                    'WELY' .. elastic section modulus of a member (major axis)
                    'WELZ' .. elastic section modulus of a member (minor axis)
                    'WPLY' .. plastic section modulus of a member (major axis)
                    'WPLZ' .. plastic section modulus of a member (minor axis)
                    'PROFILE' .. choosing a profile from a given catalogue
                    'BF' .. width of flange
                    'TF' .. thickness of flange
                    'H' .. height of the profile
                    'TW' .. thickness of the web
        """

        self.name = name
        self.value = value
        self.lb = lb
        self.ub = ub

        if any([target_property,target_properties, target_objects, target_object]):

                if not any([target_properties, target_property]):
                    raise ValueError("target_property/properties cannot be None "
                                     "if target_object/objects is defined!")

                if not any([target_objects, target_object]):
                    raise ValueError("target_object/objects cannot be None "
                                     "if target_property/properties is defined!")

                if target_properties is None and target_property is not None:
                    target_properties = [target_property]
                if target_objects is None:
                    target_objects = [target_object]

                target = {"property": target_properties,
                          "objects": target_objects}

        self.target = target
        self.profiles = profiles
        self.locked = False
        self.branch_priority = 0
        self.scaling = 1.0
        self.target_fun = target_fun

        # Lagrange multiplier for lower bound
        self.lb_mult = 1.0
        # Lagrange multiplier for upper bound
        self.ub_mult = 1.0

    def __repr__(self):

        # return self.name + ": [" + str(self.lb) + "," + str(self.ub) + "]"
        if self.locked:
            fixed = " (fixed to {0:g})".format(self.value)
        else:
            fixed = ""

        return str(self.lb) + " <= " + self.name + " <= " + str(self.ub) + fixed

    def lock(self, val=None):
        """
        Locks the variable, so it's value can't be changed

        Parameters:
        -----------
        :param val: value to give to variable before locking
        """
        if val is not None:
            self.substitute(val)

        self.locked = True

    def unlock(self):
        """
        Unlocks the variable so it's value can be changed
        """
        self.locked = False

    def scale(self, scaling_factor):
        """ Scale variable by the factor 'scaling_factor' 
            This alters the variable bounds. In 'substitute' method, scaling
            is reflected such that the variable value is multiplied by
            'scaling_factor'
        """

        self.scaling = scaling_factor
        self.lb = self.lb / scaling_factor
        self.ub = self.ub / scaling_factor

    def substitute(self, new_value):
        """
        Changes variable value and modifies target
        :param new_value:
        :return:
        """
        if not self.locked:
            """ Substitute a new value for the variable """
            self.value = new_value
            """ Modify target object(s) if any """
            if self.target is not None:
                for obj in self.target['objects']:
                    prop = self.target['property']
                    """ If the variable affects several properties at a time
                        'prop' is a list of property names
                    """
                    if isinstance(prop, list):
                        for p in prop:
                            """ If the variable has a 'target_fun',
                                evaluate it for right properties
                            """
                            if self.target_fun is not None:
                                """ If 'p' is not a property that has a functional dependency,
                                    then do regular substitution
                                """
                                try:
                                    new_val = self.target_fun[p](new_value * self.scaling)
                                except KeyError:
                                    new_val = new_value * self.scaling

                                obj.__setattr__(p, new_val)
                            else:
                                obj.__setattr__(p, new_value)

                    elif isinstance(self, IndexVariable):
                        obj.__setattr__(prop, new_value)
                    elif isinstance(self, BinaryVariable):
                        """ If a binary variable has a target, assume that target
                            is the section to be substituted. The section is
                            given in the 'section' attribute of the variable.
                        """
                        if new_value == 1:
                            if prop == 'profile' or prop == 'cross_section':
                                obj.__setattr__(prop, self.section)
                        else:
                            """ If the target of a binary variable is a member,
                                the binary variable is a member existence variable
                                used in topology optimization. In that case,
                                mark the member as passive.
                            """
                            if prop == 'member':
                                obj.__setattr__('active', False)
                    else:
                        obj.__setattr__(prop, new_value * self.scaling)

    def discrete_violation(self, x):
        """ Identify the violation of value 'x' from allowable (Discrete)
            value. Returns 0 for continuous variables
        :return:

        """
        return 0


class IntegerVariable(Variable):
    """ Class for integer variables

        Integer variables can take integer values from the interval lb, ub
    """

    def __init__(self,
                 name: str = "",
                 lb: int = 0,
                 ub: int = 1e5,
                 **kwargs):
        """ Constructor

            Arguments:
                name .. string stating the name of the variable
                lb .. lower bound
                ub .. upper bound
        """

        super().__init__(name, lb, ub, **kwargs)

    def is_allowed_value(self, x):
        """ Check if 'x' is an integer within the range of the variable """

        if self.lb <= x <= self.ub:
            """ If x is within INT_TOL of its rounded integer value,
                it is deemed integer
            """
            if math.isclose(x, round(x, 0), abs_tol=INT_TOL):
                return True
            else:
                return False
        else:
            return False

    def discrete_violation(self, x):
        """ Identify the violation of value 'x' from allowable (Discrete)
            value. Returns 0 for continuous variables
        """
        violation = abs(x - round(x, 0))
        if violation <= INT_TOL:
            violation = 0

        return violation


class BooleanVariable(IntegerVariable):
    """ Class for boolean variables

        Boolean variables can only have values 0 or 1

    """

    def __init__(self,
                 name="",
                 **kwargs):
        """ Constructor

            Arguments:
                name .. string stating the name of the variable
        """
        super(IntegerVariable, self).__init__(name, 0, 1, **kwargs)


class BinaryVariable(IntegerVariable):
    """ Class for binary variables

        Binary variables can only have values 0 or 1

    """

    def __init__(self, name="", section=None, objects=None, **kwargs):
        """ Constructor

            Arguments:
                name .. string stating the name of the variable
        """
        super(IntegerVariable, self).__init__(name, 0, 1, **kwargs)

        self.section = section
        self.objects = objects


class IndexVariable(IntegerVariable):
    """
    Class for index variables.
    
    An index variable is usually used to select a profile for
    a member.
    
    """

    def __init__(self, name="", values=None, target=None, **kwargs):
        """ Constructor:
            :param values: list of possible 
        """
        self.values = values

        if values is not None:
            lb = 0
            ub = len(values) - 1
        else:
            lb = None
            ub = None

        super().__init__(name, lb, ub, target=target, **kwargs)

    def substitute(self, new_value):
        """ Substitute the profile corresponding to the integer new_value
        """
        try:
            # Take into account that new_value can be a float
            new_value = int(new_value)
            super().substitute(self.values[new_value])
        except:
            raise ValueError(
                f"Input {new_value} is erroneous "
                "IndexVariable's value must be either"
                " index or a value from the given list!\n"
                f"Values: {self.values}")

        self.value = new_value

    # @property
    # def idx(self):
    #     if self.value in self.values:
    #         return self.values.index(self.value)
    #     return self.value


class DiscreteVariable(Variable):
    """ Class for general discrete variables """

    def __init__(self, name="", values=None, target=None, **kwargs):
        """ Constructor

            Parameters:
            -----------
                :param name: string stating the name of the variable
                :param values: list of allowable discrete values

                Variables:
                ----------
                :ivar values: list of allowable discrete values
                :ivar name: name of the variable (string)
        """

        self.values = values

        if values is not None:
            lb = min(values)
            ub = max(values)

        else:
            lb = None
            ub = None

        super().__init__(name, lb, ub, target=target, **kwargs)

    def is_allowed_value(self, x):
        """ Check if 'x' is one of the values listed in
            self.values
        """

        if self.lb <= x <= self.ub:
            """ If x is within INT_TOL of any of the allowed values,
                it is deemed allowed
            """
            if min([abs(val - x) for val in self.values]) < INT_TOL:
                return True

            """
            for val in self.values:
                if abs(x-val) < INT_TOL:            
                    return True        
            """
            return False
        else:
            return False

    def discrete_violation(self, x):
        """ Identify the violation of value 'x' from allowable (Discrete)
            value.
        """

        violation = min([abs(val - x) for val in self.values])

        if violation <= DISC_TOL:
            violation = 0

        return violation

    def smaller_discrete_value(self, x):
        """ Returns discrete value smaller than 'x' and closest to it  """

        return max(list(filter(lambda val: (val - x < 0), self.values)))

        # diff = [val-x for val in self.values]

    def larger_discrete_value(self, x):
        """ Returns discrete value smaller than 'x' and closest to it  """

        return min(list(filter(lambda val: (val - x > 0), self.values)))


class CrossSectionVariable(Variable):
    """ Class for cross-sectional variables (continuous) """

    def __init__(self, name, lb=XLB, ub=XUB, target=None, profiles=None, value=None, target_fun=None):

        super().__init__(name, lb, ub, target, profiles, value, target_fun=target_fun)

    def substitute(self, new_value):
        """
        Changes variable value and modifies target
        :param new_value:
        :return:
        """
        if not self.locked:
            """ Substitute a new value for the variable """
            self.value = new_value

            for obj in self.target['objects']:
                prop = self.target['property']
                """ If the variable affects several properties at a time
                    'prop' is a list of property names
                """
                if isinstance(prop, list):
                    for p in prop:
                        """ If the variable has a 'target_fun',
                            evaluate it for right properties
                        """
                        if self.target_fun is not None:
                            """ If 'p' is not a property that has a functional dependency,
                                then do regular substitution
                            """
                            try:
                                new_val = self.target_fun[p](new_value * self.scaling)
                            except KeyError:
                                new_val = new_value * self.scaling

                            obj.__setattr__(p, new_val)
                        else:
                            obj.__setattr__(p, new_value)
                else:
                    obj.cross_section.__setattr__(prop, new_value * self.scaling)


if __name__ == "__main__":
    cont_var = Variable(
        name="Jatkuva muuttuja",
        lb=0,
        ub=100,
        value=50
    )
    print(cont_var)
    print(cont_var.value)
    cont_var.substitute(75.123456)
    print(cont_var.value)




# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
""" Roof truss class
    Classes:
        RoofTruss -- subclass of Truss (see truss.py)

    Written by Kristo Mela
    18.4.2017
"""

import numpy as np
import math

from metku.structures.truss import Truss, TrussMember, TrussNode
#from truss import TrussMember
#import truss as tr
from metku.sections.steel.RHS import SHS

top_node_label = "TOP"
bottom_node_label = "BOTTOM"
    
top_member_label = "TC"
bottom_member_label = "BC"
brace_label = "BR"

class RoofTruss(Truss):
    """ Class for single-span steel roof trusses
        Subclass of 'Truss'
        Uses classes:
        'Cross-Section' (and its subclasses)
        'RHSJoint' (and its subclasses)
        'FrameFEM'
        'SteelMember'
        
        Attributes:
            span -- truss span
            h1   -- height from bottom chord to support
            slope -- top chord slope (e.g. slope = 1/20)
            braces -- list of brace members
            top_chord -- list of top chord members
            bottom_chord -- list of bottom chord members
            joints -- list of joints
            column -- column profile (optional)
            fyt -- top chord yield strength
            fyb -- bottom chord yield strength
            fy -- brace yield strength
        
        Methods:
            top_chord_angle -- determine top chord angle
            nbraces -- number of braces
            add_roof_member -- adds roof truss member
            roof_member_coord -- get roof member coordinates
            roof_truss_nodes -- create roof truss nodes
            top_chord_direction -- top chord direction vector
            brace_angle_top_chord -- angle between brace and top chord
            brace_angle_bottom_chord
    """
    
    top_node_label = "top"
    bottom_node_label = "bottom"
    
    top_member_label = "top"
    bottom_member_label = "bottom"
    brace_label = "brace"
    
    def __init__(self,span,h1,slope,topology="K",divX=4):
        """ Constructor
            
            NOTE: it might be better to leave creation of members and nodes
            out of the constructor in order to better allow free topologies
            
             span .. span of the truss (user defines units)
             h1 .. height from bottom chord to support
             slope .. slope of the top chord (e.g. slope = 1/20).
             members .. dictionary of truss members. The key:value pair is
                 member_id:member_object, where member_id is a string with
                 the unique identifier of the member and member_object is the
                 actual member object.
                 
                 the member_id identifies a member as a top chord member,
                 bottom chord member or a brace.
            
        """
        Truss.__init__(self)
        self.span = span
        self.h1 = h1
        # self.h2 = h2
        self.slope = slope
        
        
        self.members = {}
        
        self.braces = []
        self.top_chord = Chord()
        self.bottom_chord = Chord()
        self.joints = []
        self.column = 0.0
        self.fyt = 355.0
        self.fyb = 355.0
        self.fy = 355.0
        
        self.fem = []
    
    #THIS STUFF CAN BE LEFT OUT OF THE CONSTRUCTOR!
        # Create nodal coordinates
        self.roof_truss_nodes(divX,topology)
        
        # RETHINK these attributes!
        # Truss has 'nodes' now instead of 'x' and 'xmod'
        # 'nodes' might also include 'xmod' information!        
        self.roof_truss_members(topology)

        #self.symmetry_nodes()              
        #self.symmetry_pairs()        
        
        #self.determine_joints()
        
    
    def top_chord_angle(self):
        """ Angle of the top chord with respect to the horziontal axis. """    
        a = math.degrees(math.atan(self.slope))
        return a
   
    def nbraces(self):
        """ Number of brace members """
        NB = len(self.braces)
        return NB
    
    def add_node(self,x,y,chord):
        """ Create node and add it to the list of nodes
            
            chord -- "top" or "bottom"
        """
        newNode = RoofTrussNode(x,y,chord)
        self.nodes.append(newNode)
        
        if chord == "top":
            self.top_chord.add_node(newNode)
        else:
            self.bottom_chord.add_node(newNode)    
    
    def add_member(self,n1,n2,profile,member_type):
        """  Add new member 
        
            Calls the AddMember method of class 'Truss', with additional
            information on member type
        """        
        if member_type == "brace":
            mNew = Brace(n1,n2,profile)
            self.braces.append(mNew)
        elif member_type == "top":
            mNew = TrussMember([n1,n2],profile)
            self.top_chord.members.append(mNew)
        else:
            mNew = TrussMember([n1,n2],profile)
            self.bottom_chord.members.append(mNew)
    
        self.members.append(mNew)
    
    def roof_truss_nodes(self,n,topology):
        """ Create roof truss nodes for given topology and division
            n -- number of divisions on one half of the truss
            topology -- one of the standard topologies, 'K', 'N', or KT
        """
        L = self.span
        h = self.h1
        k = self.slope
        x = 0.0
        dx = 0.5*L/n
        
        """ First (left-most) supported node """
        self.add_node(0.0,0.0,top_node_label)
        
        if topology == "N":            
            for i in range(n):
                x = x+dx
                self.add_node(x,-h,bottom_node_label)
                self.add_node(x,k*x,top_node_label)
            
            for i in range(n-1):
                x = x+dx
                self.add_node(x,-h,self.bottom_node_label)
                self.add_node(x,-k*(x-L),self.top_node_label)

            self.add_node(L,0,top_node_label)
            
        elif topology == "K":
            for i in range(n):
                x = x+dx
                self.add_node(x-0.5*dx,-h,bottom_node_label)
                self.add_node(x,k*x,top_node_label)
            
            for i in range(n-1):
                x = x+dx
                self.add_node(x-0.5*dx,-h,bottom_node_label)
                self.add_node(x,-k*(x-L),top_node_label)

            self.add_node(L-0.5*dx,-h,bottom_node_label)
            self.add_node(L,0,top_node_label)
                
        elif topology == "KT":
            for i in range(n):
                x = x+dx
                self.add_node(x-0.5*dx,k*(x-0.5*dx),top_node_label)
                self.add_node(x-0.5*dx,-h,bottom_node_label)
                self.add_node(x,k*x,top_node_label)
            
            for i in range(n-1):
                x = x+dx
                self.add_node(x-0.5*dx,-k*(x-0.5*dx-L),top_node_label)
                self.add_node(x-0.5*dx,-h,bottom_node_label)
                self.add_node(x,-k*(x-L),top_node_label)

            self.add_node(L-0.5*dx,-k*(-0.5*dx),top_node_label)
            self.add_node(L-0.5*dx,-h,bottom_node_label)
            self.add_node(L,0,top_node_label)
    
    def roof_truss_members(self,topology):
        """ Creates members based on given nodal coordinates """        

        """ Default profiles """
        defTop = SHS(150,6)
        defBottom = SHS(140,5)
        defBrace = SHS(60,4)

        N = self.nnodes()

        if topology == "N":
            half_nodes = int(N/2)
            for i in range(half_nodes):
                if self.nodes[i].chord == top_node_label:
                    """ node is located at the top chord """
                    n1 = i+2
                    self.add_member(i,n1,defTop,top_member_label)
                    n2 = i+1
                    self.add_member(i,n2,defBrace,brace_label)
                else:
                    """ node is located at the bottom chord """
                    n1 = i+1
                    self.add_member(i,n1,defBrace,brace_label)
                    n2 = i+2
                    self.add_member(i,n2,defBottom,bottom_member_label)

            self.add_member(half_nodes-1,half_nodes+2,defBottom,bottom_member_label)

            for i in range(int(N/2),int(N-3)):
                if self.nodes[i].chord == top_node_label:
                    # node is located  at the top chord                                
                    n1 = i+2
                    self.add_member(i,n1,defTop,top_member_label)
                else:
                    # node is located at the bottom chord        
                    n1 = i+1
                    self.add_member(i,n1,defBrace,brace_label)
                    n2 = i+2
                    self.add_member(i,n2,defBottom,bottom_member_label)
                    n3 = i+3
                    self.add_member(i,n3,defBrace,brace_label)

            # Final two nodes:
            # 1) Last node of the bottom chord
            n1 = N-2
            n2 = N-1
            self.add_member(N-3,n1,defBrace,brace_label)
            self.add_member(N-3,n2,defBrace,brace_label)

            # 2) Last node of the top chord
            n1 = N-1
            self.add_member(N-2,n1,defTop,top_member_label)
        elif topology == "K":
            for i in range(N-2):
                if self.nodes[i].chord == top_node_label:
                    # node is located  at the top chord        
                    n1 = i+2
                    n2 = i+1
                    self.add_member(i,n1,defTop,top_member_label)
                    self.add_member(i,n2,defBrace,brace_label)
                else:
                    # node is located at the bottom chord
                    n1 = i+1
                    self.add_member(i,n1,defBrace,brace_label)
                    n2 = i+2
                    self.add_member(i,n2,defBottom,bottom_member_label)
        
            self.add_member(N-2,N-1,defBrace,bottom_member_label)
        
            
        elif topology == "KT":
            for i in range(N-1):
                if self.nodes[i].type == top_node_label:
                    self.add_member(i,i+1,defTop,top_member_label)
                    self.add_member(i,i+2,defTop,top_member_label)
                else:
                    """ bottom chord node """
                    self.add_member(i,i+1,defBrace,brace_label)
                    if i+3 < N-1:
                        self.add_member(i,i+3,defBottom,bottom_member_label)
                            

    def top_chord_direction(self,up_slope=True):
        """ Direction vector of the top chord:
            up_slope -- slope going upwards
        """
        a = self.top_chord_angle()
        
        ca = math.cos(math.radians(a))
        sa = math.sin(math.radians(a))
        
        if up_slope == True:
            v = np.array([ca,sa,0])
        else:
            v = np.array([ca,-sa,0])
        
        return v
    
    def brace_angle_top_chord(self,m):
        """ Calculate angle betweena brace and top chord """
        
        # Get the direction vector of member m
        #v = self.member_direction_vector(m)
        v = self.members[m].direction_vector()

        #X = self.member_coord(m)
        X = self.members[m].coord()
        aTop = self.top_chord_angle()
            
        """ if the brace is located on the left hand side
            of the ridge, then the direction vector of the
            top chord has a positive y-component.
        """
        if max(X[:,0]) <= 0.5*self.span:
            t = self.top_chord_direction(True)
        else:
            t = self.top_chord_direction(False)
                
        dvt = np.dot(v,t)
        if dvt < 0:
            a = 180-math.degrees(math.acos(dvt))
        else:
            a = math.degrees(math.acos(dvt))

        return a

    def brace_angle_bottom_chord(self,m):
        """ Determine the angle between brace and the bottom chord """

        #v = self.member_direction_vector(m)
        v = self.members[m].direction_vector()
        """ angle of the bottom chord with x-axis.
            it is assumed that the bottom chord is horizontal.
        """

        aBottom = 0
            
        """ direction vector of the bottom chord """
        t = np.array([math.cos(aBottom),math.sin(aBottom),0])
                
        dvt = np.dot(v,t)
        if dvt < 0:
            a = 180-math.degress(math.acos(dvt))
        else:
            a = math.degrees(math.acos(dvt))

        return a
"""
    %% Properties
    properties
        
        span    % span
        h1      % height from bottom chord to support
        h2      % height from bottom chord to top chord
        slope   % slope of the top chord
        UChord  % top chord members
        LChord  % bottom chord members
        Braces  % braces
        Joints  % joints, of class 'RHSJoint' (and its subclasses)
        Column = SHS(200,8) % profile of the column.
        fyt = 355 % top chord strength [MPa]
        fyb = 355 % bottom chord strength [MPa]
        fy = 355 % brace strength [MPa]
        
        % uniform loads on the chords
        TopChordLoad = 0;
        BottomChordLoad = 0;
        
        xmod % modified nodal coordinates.
        symmNode % pairs of symmetric nodes
        
        fem % finite element model of the truss
        jointFEMNodes % nodes of the fem model located at the joints
        
    end
    
    %% Methods
    
        % moves node 'n'.
        % 'value' is a 2x1 vector with new location
        % note: n may be a vector in which case 'value' must be a matrix
        def self = MoveNode(self,n,value)            
            self.xmod(:,n) = value;
            nsymm = [];
            
            for i = 1:length(n)
                [row,col] = find(self.symmNode == n(i));
                if ~isempty(row)
                    nnew = self.symmNode(row,setdiff(1:2,col));
                    self.xmod(:,nnew) =  value(:,i);
                    if value(1,i) < 0.5*self.span
                        self.xmod(1,nnew) = self.span-self.xmod(1,nnew);
                    else
                        self.xmod(1,nnew) = self.span-self.xmod(1,nnew);
                    end
                    nsymm = [nsymm,nnew];
                end
            end
            
            n = [n,nsymm];
            modNodes = length(n);
            for r = 1:modNodes
                % update the angles of the members
                % first, detect which braces are connected to the node:
                b = self.Joints{n(r)}.Braces;
                % then, compute the new angles of the braces:            
                for i = 1:length(b)
                    % find the nodes of member b(i)
                    nodes = self.members(b(i)).n;
                    for j = 1:2                    
                        chord = self.ChordNode(nodes(j));
                        % Determine the angle between the brace and
                        % the chord corresponding to node 'n'
                        switch chord
                            case 'top'
                                a(i,1) = self.TopAngle(b(i));
                            case 'bottom'    
                                a(i,1) = self.BottomAngle(b(i));
                        end
                        % determine the index of brace b(i) in joint n
                        nd = find(self.Joints{nodes(j)}.Braces == b(i));
                        self.Joints{nodes(j)}.Joint.theta(nd) = a(i);
                        B = self.Joints{nodes(j)}.Braces;
                        switch length(B)
                            case 2                            
                                % Update gap
                                %self.GapBetweenBraces(B(1),B(2),chord);
                                self.Joints{nodes(j)}.Joint.gap = self.GapBetweenBraces(B(1),B(2),chord);
                            case 3
                                % Update gaps
                                for k = 1:2                                
                                    %self.GapBetweenBraces(B(k),B(3),chord);
                                    self.Joints{nodes(j)}.Joint.gap(k) = self.GapBetweenBraces(B(k),B(3),chord);
                                end
                        end
                    end
                end
            end
%             if length(b) > 1
%                 self.Joints{n}.Joint.gap = self.GapBetweenBraces(b(1),b(2),chord);
%             end
            
        end
        
        
        def self = setTopChordProfile(self,p)            
            % set the corresponding profile to all 
            for i = 1:length(self.UChord)
                self.members(self.UChord(i)).profile = p;             
            end
            nTop = self.TopChordNodes;
            for i = 1:length(nTop)
                self.Joints{nTop(i)}.Joint.chord = p;
                switch length(self.Joints{nTop(i)}.Braces)
                    case 2
                    % For K- and N-joints determine the gaps again:
                        self.Joints{nTop(i)}.Joint.gap = self.GapBetweenBraces(self.Joints{nTop(i)}.Braces(1),self.Joints{nTop(i)}.Braces(2),'top');
                    case 3
                        for j = 1:2
                            self.Joints{nTop(i)}.Joint.gap(j) = self.GapBetweenBraces(self.Joints{nTop(i)}.Braces(j),self.Joints{nTop(i)}.Braces(3),'top');
                        end
                end
            end
        end
        
        % Set the bottom chord profile to 'p'
        def self = setBottomChordProfile(self,p)
            for i = 1:length(self.LChord)
                self.members(self.LChord(i)).profile = p;                
            end
            nBottom = self.BottomChordNodes;
            for i = 1:length(nBottom)
                self.Joints{nBottom(i)}.Joint.chord = p;                
                switch length(self.Joints{nBottom(i)}.Braces)
                    case 2
                        % For K- N- and KT-joints, determine the gaps again:
                        self.Joints{nBottom(i)}.Joint.gap = self.GapBetweenBraces(self.Joints{nBottom(i)}.Braces(1),self.Joints{nBottom(i)}.Braces(2),'bottom');
                    case 3
                        for j = 1:2
                            self.Joints{nBottom(i)}.Joint.gap(j) = self.GapBetweenBraces(self.Joints{nBottom(i)}.Braces(j),self.Joints{nBottom(i)}.Braces(3),'bottom');
                        end                
                end
            end
        end
        
        % get the profile of the chord to which node 'n' belongs to.
        def [p,chord] = ChordProfile(self,n)                
            chord = self.ChordNode(n);
            switch self.ChordNode(n)
                case 'top'
                    %p = self.members(self.UChord(1)).profile;
                    p = self.TopChordProfile;
                case 'bottom'
                    %p = self.members(self.LChord(1)).profile;
                    p = self.BottomChordProfile;
            end
        end
        
        % set profile 'p' to braces in vector 'b'
        % this leads to an update of the gaps to the corresponding
        % K and N joints.
        def self = BraceProfile(self,p,b)
            nb = length(b);
            for i = 1:nb
                % set for the member
                self.members(b(i)).profile = p;
                % the profile must also be set for the
                % joints that the member is a part of
                n = self.members(b(i)).n;
                for j = 1:2
                    %n(j)
                    %length(self.Joints{n(j)}.Braces)
                    nbraces = length(self.Joints{n(j)}.Braces);
                    if nbraces == 1                        
                            self.Joints{n(j)}.Joint.braces = p;
                    else                        
                        % determine which index is used for brace b(i)                        
                        nd = find(self.Joints{n(j)}.Braces == b(i));                           
                        % set profile of the corresponding brace                        
                        self.Joints{n(j)}.Joint.braces{nd} = p;                        
                        % determine chord of the node ('top' or 'bottom')                            
                        chord = self.ChordNode(n(j));
                        if nbraces == 2
                            self.Joints{n(j)}.Joint.gap = self.GapBetweenBraces(self.Joints{n(j)}.Braces(1),self.Joints{n(j)}.Braces(2),chord);
                        else
                            for k = 1:2
                                self.Joints{n(j)}.Joint.gap(k) = self.GapBetweenBraces(self.Joints{n(j)}.Braces(k),self.Joints{n(j)}.Braces(3),chord);
                            end
                        end
                    end
                end
            end
        end                               
        
        
        % Determines the symmetry pairs of roof truss members
        % Symmetry axis is located at midspan
        def self = SymmetryPairs(self)
            E = [self.members.n];
            S = zeros(self.nE,2);
            xS = 0.5*self.span;           
            for i = 1:self.nE               
                S(i,:) = sort(abs(self.x(1,E(:,i))-xS));
            end
            for i = 1:self.nE
                if isempty(self.members(i).pair)
                    for j = i+1:self.nE
                        if norm(S(i,:)-S(j,:),2) < 1e-3
                            self.members(i).pair = j;
                            self.members(j).pair = i;
                        end
                    end
                end
            end
        end
        
        % Find pairs of nodes that lie symmetric with respect to
        % the mid span of the truss
        def self = SymmetryNodes(self)
            S = [];
            xS = 0.5*self.span;
            n1 = find(self.x(1,:) < xS);
            n2 = find(self.x(1,:) > xS);
            S1 = abs(self.x(1,n1)-xS);
            S2 = abs(self.x(1,n2)-xS);
            for i = 1:length(n1)
                npair = find(abs(S2-S1(n1(i)))<1e-3);
                switch length(npair)
                    case 2
                    %n1(i)
                    %n2(npair(abs(self.x(2,n2(npair))-self.x(2,n1(i))) < 1e-3))
                        S = [S;n1(i),n2(npair(abs(self.x(2,n2(npair))-self.x(2,n1(i))) < 1e-3))];
                    case 1                
                        S = [S;[n1(i),n2(npair)]];                        
                end
            end
            self.symmNode = S;
        end
        
        % Determine eccentricity at node 'n'
        def e = Eccentricity(self,n)
            chord = self.ChordNode(n);
            switch chord
                case 'top'                    
                    v = self.TopDirection(n);                    
                case 'bottom'                    
                    v = [1;0];                    
            end
            
            e = self.Point2Line(self.xmod(:,n),self.x(:,n),v);

            switch chord
                case 'top'
                    if self.xmod(2,n) < self.x(2,n)
                        e = -e;
                    end
                case 'bottom'
                    if self.xmod(2,n) > self.x(2,n)
                        e = -e;
                    end
            end            
        end
        
 
        
    %% Methods in separate files
    methods
        % Calculate points where braces meet the surface of the chords
        [xChord,N,xProj] = ChordSurfacePoint(self,m,verb)
        % Determine joint types and generate joint selfects
        self = DetermineJoints(self);
        % Calculate gaps between braces
        g = GapBetweenBraces(self,m1,m2,chord);
        % Determine node locations such that eccentricity is minimized
        self = MinimizeEccentricity(self,x0);
        % for analytical expressions
        [self,P,xsol] = MinimizeEccentricity2(self,x0);
        % Creates a 'FrameFEM' model of the truss
        [fem,self,XndCenter] = CreateFEModel(self);
        % Writes a str file to be read in Robot Structural analysis
        WriteRobotFile(self,filename);
        % Perform member design according to EN 1993        
        [self,UtRatio] = MemberDesign(self,verb,allClass1or2)
    end
    
    %% Static methods
    methods (Static = true)
        def R = RotationMatrix(a)
        %
        % 2D rotation matrix
        % 
        % 28.12.2016 -k.m.
        %

        ca = cosd(a);
        sa = sind(a);

        R = [ca, -sa; sa, ca];
        end
        
        % distance of point r1 from line passing through point r0
        % with direction vector v
        def d = Point2Line(r1,r0,v)
           if length(r1) == 2
               r1 = [r1;0];
               r0 = [r0;0];
               v = [v;0];
           end
           d = norm(cross(r1-r0,v),2)/norm(v,2);
        end
    end
end

"""

class Chord:
    """ Class for storing chord data and manipulating things """
    def __init__(self):
        self.members = []
        self.nodes = []
        self.profile = None
        self.fy = 355
        self.slope = 0
    def angle(self,position):
        """ chord angle """
        
    def add_node(self,newNode):
        self.nodes.append(newNode)
        
    def add_member(self,newMember):
        """ Add member to chord """
        self.members.append(newMember)

    def set_profile(self,p):
        """ Sets a new profile for all members """
        self.profile = p

        for m in self.members:
            m.profile = p

    def set_fy(self,fy):
        """ Sets a new yield strength for all members """
        self.profile = fy
        
        for m in self.members:
            m.profile.fy = fy

class Brace(TrussMember):
    """ Class for braces 
        
        Subclass of TrussMember of 'truss' module
        The main difference is that the coordinates are taken from the
        variable 'xmod' instead of 'x'
    """

    def __init__(self,n1,n2,p):
        TrussMember.__init__(self,[n1,n2],p)

    def coord(self):
        """ Nodal coordinates of brace member
        
        """
        X = np.array([self.nodes[0].xmod,self.nodes[1].xmod])
        return X

    def is_vertical(self):
        res = False
        X = self.coord()
        if abs(X[0,0]-X[1,0]) < 1e-6:
            res = True

        return res

class RoofTrussNode(TrussNode):
    
    def __init__(self,x,y,chord):
        """ RoofTrussNode
            
            Subclass of TrussNode
            x, y -- coordinates
            chord -- "top" or "bottom", denotes the chord
        """
        TrussNode.__init__(self,x,y)
        
        self.xmod = np.array([x,y,0])
        self.chord = chord

""" For testing roof trusses """
if __name__ == "__main__":
    import sys
    topology = sys.argv[1]
    divX = int(sys.argv[2])
    t = RoofTruss(24000,2200,1/20,topology,divX)
    print("Number of nodes:",t.nnodes())
    print("Number of members:",t.nmembers())
    print("Slope (degrees):",t.top_chord_angle())
    
    nodes = [0,2,3,4]
    for n in nodes:
        print(t.nodes[n].x)
        print(t.nodes[n].xmod)
        print(t.nodes[n].chord)
    """
    braces = [2,5]
    for b in braces:
        print("Angle of member", b, "with top chord:",t.brace_angle_top_chord(b))
        print("Angle of member", b, "with bottom chord:",t.brace_angle_bottom_chord(b))
    """
    t.draw()
    
class Joint:
    """ Class for truss joints 
        properties:
            init_node .. initial node, where the center lines of the
                         braces meet at the center line of the chord
            eccentricity_node .. if the structural model contains a single eccentricity element,
                                the eccentricity node is stored here
            chord_face_nodes .. if the braces in the structural model end at
                                chord surfaces, the nodes are placed here
            eccentricity_elements .. element(s) from chord center line to eccentricity node
                                     or chord face
            gap_member .. member at the gap
            braces .. brace members connected to the joint
            chord_members .. chord members connected to the joint
    
    """
    
    def __init__(self,init_node):
        
        self.init_node = init_node
        self.eccentricity_node = None
        
        self.eccentricity_elements = None
        self.chord_face_nodes = None
        self.gap_member = None

        self.braces = []
        self.chord_members = []
    
    def add_brace(self,new_brace):
        self.braces.append(new_brace)
    
    def add_chord_member(self,new_chord_member):
        self.chord_members.append(new_chord_member)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Tue Mar 26 12:20:46 2024

Class for Vierendeel trusses

@author: kmela
"""

import numpy as np
import matplotlib.pyplot as plt
from copy import copy

from metku.raami.raami import Raami
from metku.raami.raami_plane_truss import PlaneTruss, TopChord, BottomChord, TrussBrace, TubularYJoint
from metku.raami.raami_functions import divide_line_segment
from metku.raami.frame_node import FrameNode
from metku.raami.frame_member import FrameMember, SteelFrameMember, MultiSpanSteelMember, MemberGroup
from metku.raami.frame_loads import PointLoad, LineLoad, PWLineLoad, LoadIDs
from metku.raami.frame_supports import XYHingedSupport, YHingedSupport
from metku.raami.exports import AbaqusOptions

from metku.framefem import LineLoad as FEMLineLoad
from metku.framefem.elements.ebbeam import EBBeam, EBBeam3D
from metku.framefem.elements.rod import Rod
from metku.sections.steel.RHS import RHS, SHS
from metku.sections.steel.ISection import IPE
from metku.eurocodes.en1993.en1993_1_8.rhs_joints import Line, RHSKGapJoint, RHSYJoint, RHSKTGapJoint
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import full_strength_weld_tube

from metku.cost.workshop import Workshop


#from loadIDs import LoadIDs

def unit_vector(x1,x2):
    """ Creates a unit vector between points x1 and x2 """
    
    return (x2-x1)/np.linalg.norm(x2-x1)

class VierendeelTruss(PlaneTruss):

    def __init__(self, origin=(0, 0), span=10000, height=1000, panels=5,
                 num_elements=6,
                 name="Raami Vierendeel Truss",
                 workshop=Workshop()):
        """ Constructor 
            Parameters:
            ------------
            :param origin: location of the origin (tuple)
            :param span: span of the truss [mm]
            :param height: height of the truss [mm]
            :param panels: number of panels, i.e. division of span to vertical braces
            :param num_elements: number of elements per member
            
            :type origin: tuple
            :type num_elements: int
            :type fem_model: FrameFEM (optional)
    
    
            Variables:
            ----------
            :ivar truss:
            :ivar origin:
            :ivar top_chords: list of members constituting the top chord
            :ivar bottom_chords: list of bottom chord members
            :ivar braces: dict of brace members
            :ivar joints: dict of truss joints
            
            :ivar _H0: y-coordinate of the bottom chord
            :ivar _H1: height at
            :ivar _H2: height at
            :ivar _H3: height at
            :ivar L1: span
            :ivar L2: span
            
        
        """
        PlaneTruss.__init__(self,origin=origin, span=span, left_height=height,
                     mid_height=height, right_height=height, 
                     bot_chord_dx_left = 0, bot_chord_dx_right = 0,
                     rotation = 0,
                     num_elements=num_elements,
                     name="Raami Vierendeel Truss")
        
        # Braces are modelled using beam elements
        self.braces_as_beams = True
        
        self._H = height
        self._npanels = panels                
    
    def weight_detailed(self,verb=False):
        """ Returns detailed weight 
            The weight of chords and braces are separated.
        """
        Wtop = 0
        Wbottom = 0
        Wbraces = 0
        
        for mem in self.members.values():
            W = mem.weight()
            
            if isinstance(mem,TopChord):
                Wtop += W
            elif isinstance(mem,BottomChord):
                Wbottom += W
            else:
                Wbraces += W
        
        Wtot = Wtop + Wbottom + Wbraces
        
        if verb:
            print(f"Truss weight: {Wtot:.2f} kg")
            print(f"Top Chord: {Wtop:.2f} kg ({Wtop/Wtot*100:.1f}%)")
            print(f"Bottom Chord: {Wbottom:.2f} kg ({Wbottom/Wtot*100:.1f}%)")
            print(f"Braces: {Wbraces:.2f} kg ({Wbraces/Wtot*100:.1f}%)")
        
        return Wtop, Wbottom, Wbraces
    
    def generate_topology(self, 
                          top_chord_profile = RHS(160,160,5),
                          bottom_chord_profile = RHS(140,140,5),
                          brace_profile = RHS(90,90,5),
                          nel_chord=4,
                          nel_brace=4):
        """ Generates truss members for a given number of braces 
        
            :param: topology .. 'K', 'N', or 'KT'
            :param: ndiv .. number of divisions on one half of the truss
            
        """

        # Generate nodes first
            
        # Bottom chord nodes:            
        dx = self.span/self._npanels  
        ycoord = self.origin[1]
        
        S = np.linspace(0,1,self._npanels+1)

        X0 = np.array([self.origin[0],ycoord])
        X1 = np.array([self.origin[0]+self.span,ycoord])  
        Xbot = [FrameNode(X0+s*(X1-X0)) for s in S]
            
        
        # Top chord nodes:
        # Local coordinates of the top chord
                        
        X0 = np.array([self.origin[0],ycoord+self._H])
        X1 = np.array([self.origin[0]+self.span,ycoord+self._H])
          
        Xtop = [FrameNode(X0+s*(X1-X0)) for s in S]                      
        
        print(Xtop)
        for Xt in Xtop:
            self.add(Xt) 
                        
        for Xb in Xbot:
            self.add(Xb)    
            
            
        # Generate top chord members
        for i in range(len(Xtop)-1):                
            self.add(TopChord([Xtop[i],Xtop[i+1]],copy(top_chord_profile),nel=nel_chord))
        
        # Generate bottom chord members
        for i in range(len(Xbot)-1):
            self.add(BottomChord([Xbot[i],Xbot[i+1]],copy(bottom_chord_profile),nel=nel_chord))
            
        # Generate braces
        for X, Y in zip(Xtop,Xbot):
            self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
        
        for brace in self.braces.values():
            brace.hinges = [False,False]            
        
        self.top_nodes = Xtop
        self.bottom_nodes = Xbot
            
    def generate_supports(self):
        """ Creates supports to the truss """
        
        self.add(XYHingedSupport(self.bottom_nodes[0]))        
        self.add(YHingedSupport(self.bottom_nodes[-1]))
    
    def generate_joints(self):
        """ Creates joints for the truss. These are welded tubular joints by default. """
        
        joints = {}
        
        # First, determine which members are connected to each node
        # This information is stored in 'joints' dict.
        for node in self.nodes:
            new_joint = {'chord':[],'braces':[]}
            for mem in self.top_chord:
                if node in mem.nodes:
                    new_joint['chord'].append(mem)
                    if len(new_joint['chord'])==2:
                        break                    
            
            for mem in self.bottom_chord:
                if node in mem.nodes:
                    new_joint['chord'].append(mem)
                    if len(new_joint['chord'])==2:
                        break
            
            for mem in self.braces.values():
                if node in mem.nodes:
                    new_joint['braces'].append(mem)
                    # The brace knows to which joints it belongs.
                    if node in self.top_nodes:
                        mem.top_joint = new_joint
                    else:
                        mem.bottom_joint = new_joint
            
            
            joints[node] = new_joint
        
        # Create joints. As a first guess, the joint type is
        # based on number of braces. This may not be accurate for
        # determining X joints, because both X and K joints have two braces.
        # The user must probably provide some additional information here to
        # distinguish the proper joint type, or the type will be based on
        # the results of structural analysis.
        njoint = 0
        for node, joint in joints.items():
            
            nbraces = len(joint['braces'])
            
            if nbraces == 1:                
                newJoint = TubularYJoint(node, joint['chord'], joint['braces'])
                self.add(newJoint)
                njoint += 1
            elif nbraces == 0:
                # Number of braces can be 0, if top or bottom chord is supported
                # at the node
                newJoint = None
            
            # Attach the joint to the corresponding braces
            for brace in joint['braces']:
                if node in self.top_nodes:
                    brace.top_joint = newJoint
                else:
                    brace.bottom_joint = newJoint
            
            # Attach the joint to the corresponding chord members
            if not newJoint is None:
                for chord in joint['chord']:                
                    chord.joints.append(newJoint)
        
        # Set the actual geometry for all joints
        for joint in self.joints.values():  
            angle = joint.brace_angle(joint.braces[0],geometry="actual")
            joint.joint.angle = angle
    
    def generate_uniform_load(self,q,chord="top",load_id=LoadIDs['ULS'],ltype='live'):
        """ Creates a uniform line load to one of the chords """

        vals = [q,q]

        if chord == "top":
            load_name = 'TC_' + ltype + '_line'
            for mem in self.top_chord:
                self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name, "global"))
        elif chord == "bottom":          
            load_name = 'BC_' + ltype + '_line'
            for mem in self.bottom_chord:
                self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name, "global"))
    
    def generate_chord_point_load(self,p,chord="top",load_id=LoadIDs['ULS'],ltype='live'):
        """ Creates point loads to one of the chords """

        #(self, node, v, load_id=LoadIDs['ULS'], ltype='live',
        #name='PointLoad',f=1.0)

        if chord == "top":
            load_name = 'TC_' + ltype + '_point'
            self.add(PointLoad(self.top_nodes[0], [0.0,p[0],0.0], load_id, ltype, load_name + '_1',1.0))
            self.add(PointLoad(self.top_nodes[-1], [0.0,p[0],0.0], load_id,ltype, load_name + '_1',1.0))
            for node in self.top_nodes[1:-1]:                
                self.add(PointLoad(node, [0.0,p[1],0.0], load_id, ltype, load_name + '_2',1.0))
        elif chord == "bottom":          
            load_name = 'BC_' + ltype + '_point'
            # It is assumed that the first and last bottom chord nodes are vertically supported,
            # so no vertical load is inserted there
            for node in self.bottom_nodes[1:-1]:                
                self.add(PointLoad(node, [0.0,p[1],0.0], load_id, ltype, load_name + '_2', 1.0))
    
    def symmetry(self,axis='y',sym_value=None):
        
        # Make general actions for symmetry
        super().symmetry(axis,sym_value)
        
        # Symmetry of joints
        #joint_list = list(self.joints.values())
        for joint in self.joints.values():
            #print(f"Symmetry for joint {joint.joint_id}.")
            sym_node = joint.symmetry_joint(self.joints)
            #print(sym_node)
    
    def clear_fem(self):
        """ Clears FEM data """
        super().clear_fem()
        
        for joint in self.joints.values():
            joint.clear_fem()
            
    def generate_fem(self,model="no_eccentricity",nodes_and_elements_only=False):
        """
        Creates FEM model

        Parameters
        ----------
        model : string, optional
            Type of FEM model. Allowed values are:
                "simple": all members are pin-jointed bars
                "no_eccentricity": braces meet at originally created nodes
                "en1993": for K joints, the braces meet at the intersection point of their
                        centerlines, and an eccentricity element is added
                "braces2chord_faces": braces end at chord faces.
                The default is "no_eccentricity".

        Returns
        -------
        None.

        """
        
        if model == "no_eccentricity":
            # Here, we can use the fem model generation of Raami class
            super().generate_fem()
            
            for joint in self.joints.values():
                # Generate fem nodes for the joints
                x = joint.node.coords
                for node in self.fem.nodes:
                    if np.linalg.norm(node.coord-x) < 1e-6:
                        joint.fem_nodes["xc"] = node
                        break
                            
        elif model == "ecc_elements":
            # Eccentricity elements are added for K joints
            # from the intersection of brace centerline and chord
            # face to the chord centerline.
            for joint in self.joints.values():
                # For Y joints, no eccentricity elements are used.
                newNode = self.fem.add_node(joint.node.x,joint.node.y)                    
                joint.fem_nodes['xc'] = newNode
                joint.node.fem_node = newNode
                
              
                                    
            for mem in self.braces.values():
                #Generate member elements                
                mem.generate_elements(self.fem,model,self.braces_as_beams)
                                    
            for mem in self.top_chord:
                mem.generate_elements(self.fem,model)
        
            for mem in self.bottom_chord:
                mem.generate_elements(self.fem,model)
            
       
            if not nodes_and_elements_only:
                # Generate supports
                for supp in self.supports.values():
                    supp.add_support(self.fem)
                                
                # Generate loads
                for load in self.loads.values():
                    # add_load addes the loads on the target member
                    load.add_load(self.fem)
                    # Creates new loadcase if one with same load_id does not exist
                    lcase_ids = [lc.load for lc in self.fem.loadcases.values()]
                    if load.load_id not in lcase_ids:
                        self.fem.add_loadcase(supp_id=1,load_id=load.load_id)
                
                # Set nodal degrees of freedom
                self.fem.nodal_dofs()
                
        else:
            print("Weird FEM model.")
    
    def optimize_members(self, prof_type="CURRENT",verb=False,**kwargs):    
        """ Finds minimum weight profiles for members """
               
        top = {'material': 'S355', 'class': 2, 'utility': 1.0, 'bmin':10, 'bmax':1e5}
        bottom = {'material': 'S355', 'class': 2, 'utility': 1.0,'bmin':10, 'bmax':1e5}
        braces = {'material': 'S355', 'class': 2, 'utility_tens': 1.0, 'utility_comp':1.0, 'bmin':10, 'bmax':1e5}
            
        limit_width = False
        
        for key, value in kwargs.items():
            if key == 'top':
                top = value
            elif key == 'bottom':
                bottom = value
            elif key == 'braces':
                braces = value
            elif key == 'limit_width':
                limit_width = value
        
        kmax = 10
        k = 1
                        
        while k < kmax:            
            explored = []
            MEM_PROFILES_CHANGED = []
            self.structural_analysis('all','REM')
            if verb:
                print(f"Optimize profiles, iteration {k}")
            
            if limit_width:
                b_chord_min = self.max_brace_width()
                #b_brace_max = min(self.top_chord_profile.b,self.bottom_chord_profile.b)
                b_brace_max = 1e5
                b_brace_min = 10
            else:
                b_chord_min = 10
                b_brace_max = 1e5
                b_brace_min = 10
            
            # Go through groups first
            for name, group in self.member_groups.items():
                if name == "top_chord":
                    material = top['material']
                    sec_class = top['class']
                    max_utility = top['utility']                    
                elif name == 'bottom_chord':
                    material = bottom['material']
                    sec_class = bottom['class']
                    max_utility = bottom['utility']                    
                    
                group.optimum_design(prof_type,verb,material,sec_class,max_utility,b_chord_min)
                
                for mem in group.members:
                    explored.append(mem)
                
            # Braces
            for member in self.members.values():
                #print(member)                
                if not member in explored:
                    if verb:
                        print(member)
                    explored.append(member)
                    if isinstance(member,TrussBrace):
                        material = braces['material']
                        sec_class = braces['class']
                        if list(member.NEd.values())[0] >= 0.0:
                            max_utility = braces['utility_tens']
                        else:
                            max_utility = braces['utility_comp']
                        
                        # Minimum brace width is taken from the geometry condition of T-joints:
                        # bi >= 0.25*b0
                        if limit_width:
                            b_brace_min = 0.25*max(member.top_joint.b0,member.bottom_joint.b0)
                            b_brace_max = min(member.top_joint.b0,member.bottom_joint.b0)                        
                        
                    else:
                        material = 'S355'
                        sec_class = 2
                        max_utility = 1.0
                        
                    MEM_PROFILES_CHANGED.append(member.optimum_design(prof_type,verb,material,sec_class,max_utility,bmin=b_brace_min,bmax=b_brace_max))
                                        
                    if not member.symmetry_pair is None:
                        explored.append(member.symmetry_pair)
                    
                    #print(member.cross_section)
            
            # If none of the members changed profile, exit the loop
            if not any(MEM_PROFILES_CHANGED):
                break
                
            k += 1
        
        if verb:
            print(f"Number of iterations: {k}.")
            
class SlopedTruss(PlaneTruss):
    """ Class for roof truss with a slope """
    
    def __init__(self,L1=10000,L2=10000,dx1=500,dx2=500,**kwargs):
        """
        Constructor

        Parameters
        ----------
        L1 : TYPE, optional
            DESCRIPTION. The default is 10000.
        L2 : TYPE, optional
            DESCRIPTION. The default is 10000.
        dx1 : TYPE, optional
            DESCRIPTION. The default is 500.
        dx2 : TYPE, optional
            DESCRIPTION. The default is 500.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.L1 = L1
        self.L2 = L2
        
        span = L1 + L2
        
        
        # The shape of the truss can be given by providing two of the 
        # following three dimensions. The third will be 
        h1 = 0
        h2 = 0
        self.slope = 0
        
        
        for key, value in kwargs.items():
            if key == "h1":
                h1 = value
            elif key == "h2":
                h2 = value
            elif key == "slope":
                self.slope = value
            #else:
            #    raise ValueError(f"Erroneous key {key}.")
        
        if h1 == 0:
            # User has given the height at ridge and slope
            dy = self.height_of_slope_left()
            h1 = h2-dy
        elif h2 == 0:
            # User has given the height from bottom chord to left
            # end of top chord and slope
            dy = self.height_of_slope_left()
            h2 = h1+dy
        else:
            # User has given the two heights, so slope must be
            # calculated
            dy = h2-h1
            self.slope = np.degrees(np.arctan(dy/self.L1))
            
        
        PlaneTruss.__init__(self,origin=(0, 0), span=span, left_height=height,
                     mid_height=height, right_height=height, 
                     bot_chord_dx_left = dx1, bot_chord_dx_right = dx2,
                     rotation = 0,
                     num_elements=6,
                     name="Raami Sloped Roof Truss")
        
        # Joint at the ridge
        self.ridge_joint = None
    
    def height_of_slope_left(self):
        """ Height of the truss above the origin """
        return self.L1*np.tan(np.radians(self.slope))
    
    def weight_detailed(self,verb=False):
        """ Returns detailed weight 
            The weight of chords and braces are separated.
        """
        Wtop = 0
        Wbottom = 0
        Wbraces = 0
        
        for mem in self.members.values():
            W = mem.weight()
            
            if isinstance(mem,TopChord):
                Wtop += W
            elif isinstance(mem,BottomChord):
                Wbottom += W
            else:
                Wbraces += W
        
        Wtot = Wtop + Wbottom + Wbraces
        
        if verb:
            print(f"Truss weight: {Wtot:.2f} kg")
            print(f"Top Chord: {Wtop:.2f} kg ({Wtop/Wtot*100:.1f}%)")
            print(f"Bottom Chord: {Wbottom:.2f} kg ({Wbottom/Wtot*100:.1f}%)")
            print(f"Braces: {Wbraces:.2f} kg ({Wbraces/Wtot*100:.1f}%)")
        
        return Wtop, Wbottom, Wbraces
            
        
    def generate_topology(self,topology='K',ndiv=4, 
                          top_chord_profile = RHS(160,160,5),
                          bottom_chord_profile = RHS(140,140,5),
                          brace_profile = RHS(90,90,5),
                          first_diagonal_up=False,
                          edge_verticals=True,
                          nel_chord=4,
                          nel_brace=1):
        """ Generates truss members for a given topology 
        
            :param: topology .. 'K', 'N', or 'KT'
            :param: ndiv .. number of divisions on one half of the truss
            
        """
        
        
        if topology == 'K':
            # Generate nodes first
            # Bottom chord nodes:            
            dx1 = self._dx_left            
            dx2 = self._dx_right
            ycoord = self.origin[1]-self.Hleft
            
            if dx1 > 0:
                # If the first bottom chord node is not at x = 0, do the
                # following:
                # 1) divide half of the span into ndiv trios of two diagonals
                #   and a vertical.
                # 2) each trio is symmterical with respect to the vertical.
                # 3) for the first trio, the distance in x direction between
                #    nodes is dx1 and for the others it is Dx
                S = np.linspace(0,1,ndiv)
                # Distance between nodes in x direction
                #Dx = (self.L1-2*dx1)/(ndiv-1)
                Dx = (self.L1-dx1)/(ndiv-0.5)
            else:
                S = np.linspace(0,1,ndiv+1)
                if not first_diagonal_up:
                    S = np.linspace(0,1,ndiv)
                    Xbot0 = FrameNode(np.array([self.origin[0],ycoord]))
                    Dx = self.L1/(ndiv-0.5)
                else:
                    # Distance between nodes in x direction
                    Dx = self.L1/ndiv
            
            # First half
            if dx1 == 0 and not first_diagonal_up:
                X0 = np.array([self.origin[0]+0.5*Dx,ycoord])
                X2 = np.array([self.origin[0]+self.span-0.5*Dx,ycoord])
            else:
                X0 = np.array([self.origin[0]+dx1,ycoord])
                X2 = np.array([self.origin[0]+self.span-dx2,ycoord])
            X1 = np.array([self.origin[0]+self.L1-0.5*Dx,ycoord])  
            
            Xbot = [FrameNode(X0+s*(X1-X0)) for s in S]
        
            # Second half
            X1[0] += Dx
            
            Xbot += [FrameNode(X1+s*(X2-X1)) for s in S]
        
            if dx1 == 0 and not first_diagonal_up:
                Xbot.insert(0,Xbot0)
                Xbot.append(FrameNode(np.array([self.origin[0]+self.span,ycoord])))
        
            # Top chord nodes:
            # Local coordinates of the top chord
            #            
            X0 = np.array([self.origin[0],self.origin[1]])
            X2 = np.array([self.origin[0]+self.span,self.origin[1]])
                                 
            X1 = self.xmid()
            #X2 = self.xend()
            
            v1 = X1-X0
            v2 = X2-X1
            
            #Xtop = []
            
            if dx1 == 0:
                if first_diagonal_up:
                    Xtop = [FrameNode(X0)]
                    v0 = v1/np.linalg.norm(v1)
                    top_line = Line(v=v0,p1=X0)
                    
                    for Xb2, Xb1 in zip(Xbot[:ndiv+1],Xbot[1:ndiv+2]):
                        X = np.array([0.5*(Xb2.x+Xb1.x),Xb2.y])
                        bottom_line = Line(v=np.array([0.0,1.0]),p1=X)                
                        t, Xtop1 = top_line.intersect(bottom_line)
                        
                        #print(top_line, bottom_line)                    
                        Xtop.append(FrameNode(Xtop1))
                    
                    v0 = v2/np.linalg.norm(v2)
                    top_line = Line(v=v0,p1=Xtop[-1].coords)
                    
                    for Xb1, Xb2 in zip(Xbot[ndiv+1:],Xbot[ndiv+2:]):
                        X = np.array([0.5*(Xb2.x+Xb1.x),Xb2.y])
                        bottom_line = Line(v=np.array([0.0,1.0]),p1=X)                
                        t, Xtop1 = top_line.intersect(bottom_line)
                        
                        #print(top_line, bottom_line)                    
                        Xtop.append(FrameNode(Xtop1))
                    
                    Xtop.append(FrameNode(X2))
                    
                else:
                    S = np.linspace(0,1,ndiv+1)
                
                    Xtop = [FrameNode(X0+s*v1) for s in S]
                    Xtop += [FrameNode(X1+s*v2) for s in S[1:]]
                    
            else:
                Xtop = [FrameNode(X0)]
                v0 = v1/np.linalg.norm(v1)
                top_line = Line(v=v0,p1=X0)
                
                for Xb2, Xb1 in zip(Xbot[:ndiv],Xbot[1:ndiv+1]):
                    X = np.array([0.5*(Xb2.x+Xb1.x),Xb2.y])
                    bottom_line = Line(v=np.array([0.0,1.0]),p1=X)                
                    t, Xtop1 = top_line.intersect(bottom_line)
                    
                    #print(top_line, bottom_line)                    
                    Xtop.append(FrameNode(Xtop1))
                
                v0 = v2/np.linalg.norm(v2)
                top_line = Line(v=v0,p1=Xtop[-1].coords)
                
                for Xb1, Xb2 in zip(Xbot[ndiv:],Xbot[ndiv+1:]):
                    X = np.array([0.5*(Xb2.x+Xb1.x),Xb2.y])
                    bottom_line = Line(v=np.array([0.0,1.0]),p1=X)                
                    t, Xtop1 = top_line.intersect(bottom_line)
                    
                    #print(top_line, bottom_line)                    
                    Xtop.append(FrameNode(Xtop1))
                
                Xtop.append(FrameNode(X2))
            
            for Xt in Xtop:
                self.add(Xt) 
                        
            for Xb in Xbot:
                self.add(Xb) 
                                    
        
            # Generate top chord members
            for i in range(len(Xtop)-1):                
                self.add(TopChord([Xtop[i],Xtop[i+1]],copy(top_chord_profile),nel=nel_chord))
        
            # Generate bottom chord members
            for i in range(len(Xbot)-1):
                self.add(BottomChord([Xbot[i],Xbot[i+1]],copy(bottom_chord_profile),nel=nel_chord))
        
            
            # Generate braces
            
            if dx1 == 0:
                # Diagonals
                if first_diagonal_up:
                    # Diagonals run from bottom chord to top chord
                    for i in range(1,2*ndiv+2):
                        self.add(TrussBrace([Xtop[i],Xbot[i-1]],copy(brace_profile),nel=nel_brace))
                        self.add(TrussBrace([Xtop[i],Xbot[i]],copy(brace_profile),nel=nel_brace))
                else:
                    # Diagonals run from top chord to bottom chord                    
                    for i in range(1,2*ndiv+1):
                        self.add(TrussBrace([Xbot[i],Xtop[i-1]],copy(brace_profile),nel=nel_brace))
                        self.add(TrussBrace([Xbot[i],Xtop[i]],copy(brace_profile),nel=nel_brace))
                
                # Close with verticals
                if edge_verticals:
                    self.add(TrussBrace([Xbot[0],Xtop[0]],copy(brace_profile),nel=nel_brace))
                    self.add(TrussBrace([Xbot[-1],Xtop[-1]],copy(brace_profile),nel=nel_brace))
            
            else:
                # If the first node of the bottom chord is not at the same
                # x location as the first node of the top chord,
                # the first diagonal goes from top to bottom chord.
                for i in range(0,2*ndiv):
                    self.add(TrussBrace([Xbot[i],Xtop[i]],copy(brace_profile),nel=nel_brace))
                    self.add(TrussBrace([Xbot[i],Xtop[i+1]],copy(brace_profile),nel=nel_brace))
                
                
           
        elif topology == 'N':
            # Generate nodes first
            
            # Bottom chord nodes:            
            dx1 = self._dx_left            
            dx2 = self._dx_right
            ycoord = self.origin[1]-self.Hleft
            
            if dx1 > 0:
                S = np.linspace(0,1,ndiv)
            else:
                S = np.linspace(0,1,ndiv+1)
            
            # First half
            X0 = np.array([self.origin[0]+dx1,ycoord])
            X1 = np.array([self.origin[0]+0.5*self.span,ycoord])  
            X2 = np.array([self.origin[0]+self.span-dx2,ycoord])
                        
            Xbot = [FrameNode(X0+s*(X1-X0)) for s in S]
            
            # Second half
            #X1 = np.array([self.origin[0]+0.5*self.span+dx1,ycoord])  
            Xbot += [FrameNode(X1+s*(X2-X1)) for s in S[1:]]
        
                    
            # Top chord nodes:
            # Local coordinates of the top chord
            
            
            #            
            X0 = np.array([self.origin[0],self.origin[1]])
            X2 = np.array([self.origin[0]+self.span,self.origin[1]])
                                 
            X1 = self.xmid()
            X2 = self.xend()
            
            v1 = X1-X0
            v2 = X2-X1
            
            if dx1 == 0:
                S = np.linspace(0,1,ndiv+1)
                Xtop = [FrameNode(X0+s*v1) for s in S]
                Xtop += [FrameNode(X1+s*v2) for s in S[1:]]
            else:
                S = np.linspace(0,1,ndiv)
                Xtop = [FrameNode(X0)]
                v0 = v1/np.linalg.norm(v1)
                top_line = Line(v=v0,p1=X0)
                
                for X in Xbot[:ndiv]:                    
                    bottom_line = Line(v=np.array([0.0,1.0]),p1=X.coords)                
                    t, Xtop1 = top_line.intersect(bottom_line)
                    
                    #print(top_line, bottom_line)                    
                    Xtop.append(FrameNode(Xtop1))
            
                v0 = v2/np.linalg.norm(v2)
                top_line = Line(v=v0,p1=Xtop[-1].coords)
                
                for X in Xbot[ndiv:]:
                    bottom_line = Line(v=np.array([0,1]),p1=X.coords)                
                    t, Xtop1 = top_line.intersect(bottom_line)                
                    Xtop.append(FrameNode(Xtop1))
                
                Xtop.append(FrameNode(X2))
                        
            for Xt in Xtop:
                self.add(Xt) 
                        
            for Xb in Xbot:
                self.add(Xb)    
            
            
            # Generate top chord members
            for i in range(len(Xtop)-1):                
                self.add(TopChord([Xtop[i],Xtop[i+1]],copy(top_chord_profile),nel=nel_chord))
        
            # Generate bottom chord members
            for i in range(len(Xbot)-1):
                self.add(BottomChord([Xbot[i],Xbot[i+1]],copy(bottom_chord_profile),nel=nel_chord))
            
            # Generate braces
            if dx1 == 0:
                # Verticals
                if edge_verticals:
                    for X, Y in zip(Xtop,Xbot):
                        self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                else:
                    for X, Y in zip(Xtop[1:-1],Xbot[1:-1]):
                        self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                # Diagonals
                if first_diagonal_up:
                    # Diagonals run from bottom chord to top chord
                    for X, Y in zip(Xtop[1:ndiv+1],Xbot[:ndiv]):
                        self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                        
                    for X, Y in zip(Xtop[ndiv:-1],Xbot[ndiv+1:]):
                        self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                else:
                    # Diagonals run from top chord to bottom chord
                    for X, Y in zip(Xtop[:ndiv],Xbot[1:ndiv+1]):
                        self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                        
                    for X, Y in zip(Xtop[ndiv+1:],Xbot[ndiv:-1]):
                        self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
            else:
                # Lower chord is shorter than top chord. Now the first
                # diagonal is downward anyway.
                
                # Verticals
                for X, Y in zip(Xtop[1:-1],Xbot):
                    self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                    
                # Diagonals run from top chord to bottom chord
                for X, Y in zip(Xtop[:ndiv],Xbot[:ndiv]):
                    self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                    
                for X, Y in zip(Xtop[ndiv+1:],Xbot[ndiv-1:]):
                    self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
        
        elif topology == "KT":
            # KT truss:
            # Generate nodes first
            
            # Bottom chord nodes:            
            dx1 = self._dx_left            
            dx2 = self._dx_right
            ycoord = self.origin[1]-self.Hleft
            
            if dx1 > 0:
                # If the first bottom chord node is not at x = 0, do the
                # following:
                # 1) divide half of the span into ndiv trios of two diagonals
                #   and a vertical.
                # 2) each trio is symmterical with respect to the vertical.
                # 3) for the first trio, the distance in x direction between
                #    nodes is dx1 and for the others it is Dx
                S = np.linspace(0,1,ndiv)
                # Distance between nodes in x direction
                #Dx = (self.L1-2*dx1)/(ndiv-1)
                Dx = (self.L1-dx1)/(ndiv-0.5)
            else:
                # If the first bottom chord node is not at x = 0, do the
                # following:
                # 1) divide half of the span into ndiv trios of two diagonals
                #   and a vertical.
                # 2) each trio is symmterical with respect to the vertical.
                # 3) if the first diagonal is up, the distance of it in x direction
                #    is 0.5*Dx. If the first diagonal is down, all, trios have
                #    the distance in x direction between nodes 0.5*Dx
                S = np.linspace(0,1,ndiv+1)
                if not first_diagonal_up:
                    S = np.linspace(0,1,ndiv)
                    Xbot0 = FrameNode(np.array([self.origin[0],ycoord]))
                    Dx = self.L1/ndiv
                else:
                    # Distance between nodes in x direction
                    Dx = 2*self.L1/(2*ndiv+1)
            
            # First half
            if dx1 == 0 and not first_diagonal_up:
                X0 = np.array([self.origin[0]+0.5*Dx,ycoord])
                X2 = np.array([self.origin[0]+self.span-0.5*Dx,ycoord])
            else:
                X0 = np.array([self.origin[0]+dx1,ycoord])
                X2 = np.array([self.origin[0]+self.span-dx2,ycoord])
            X1 = np.array([self.origin[0]+self.L1-0.5*Dx,ycoord])  
            
            
            Xbot = [FrameNode(X0+s*(X1-X0)) for s in S]
            
            
            
            # Second half
            X1[0] += Dx
            
            Xbot += [FrameNode(X1+s*(X2-X1)) for s in S]
        
            if dx1 == 0 and not first_diagonal_up:
                Xbot.insert(0,Xbot0)
                Xbot.append(FrameNode(np.array([self.origin[0]+self.span,ycoord])))
        
            # Top chord nodes:
            # Local coordinates of the top chord
            #            
            X0 = np.array([self.origin[0],self.origin[1]])
            X2 = np.array([self.origin[0]+self.span,self.origin[1]])
                                 
            X1 = self.xmid()
            X2 = self.xend()
            
            v1 = X1-X0
            v2 = X2-X1
            
            #Xtop = []
            
            if dx1 == 0:
                if first_diagonal_up:
                    S = np.linspace(0,1,2*ndiv+2)                    
                else:
                    S = np.linspace(0,1,2*ndiv+1)
                
                Xtop = [FrameNode(X0+s*v1) for s in S]
                Xtop += [FrameNode(X1+s*v2) for s in S[1:]]
                    
            else:
                S = np.linspace(0,1,ndiv)
                Xtop = [FrameNode(X0)]
                v0 = v1/np.linalg.norm(v1)
                top_line = Line(v=v0,p1=X0)
                
                for X in Xbot[:ndiv]:                    
                    bottom_line = Line(v=np.array([0.0,1.0]),p1=X.coords)                
                    t, Xtop1 = top_line.intersect(bottom_line)
                    
                    #print(top_line, bottom_line)                    
                    Xtop.append(FrameNode(Xtop1))
                    
                    p1New = X.coords + np.array([0.5*Dx,0])
                    bottom_line = Line(v=np.array([0.0,1.0]),p1=p1New)
                    t, Xtop2 = top_line.intersect(bottom_line)
                    
                    #print(top_line, bottom_line)                    
                    Xtop.append(FrameNode(Xtop2))
                    
                v0 = v2/np.linalg.norm(v2)
                top_line = Line(v=v0,p1=Xtop[-1].coords)
                
                for X in Xbot[ndiv:-1]:
                    bottom_line = Line(v=np.array([0,1]),p1=X.coords)                
                    t, Xtop1 = top_line.intersect(bottom_line)                
                    Xtop.append(FrameNode(Xtop1))
                    
                    p1New = X.coords + np.array([0.5*Dx,0])
                    bottom_line = Line(v=np.array([0.0,1.0]),p1=p1New)
                    t, Xtop2 = top_line.intersect(bottom_line)
                    
                    #print(top_line, bottom_line)                    
                    Xtop.append(FrameNode(Xtop2))
                
                bottom_line = Line(v=np.array([0,1]),p1=Xbot[-1].coords)                
                t, Xtop1 = top_line.intersect(bottom_line)                
                Xtop.append(FrameNode(Xtop1))
                
                Xtop.append(FrameNode(X2))
            
            for Xt in Xtop:
                self.add(Xt) 
                        
            for Xb in Xbot:
                self.add(Xb) 
                
            # Generate top chord members
            for i in range(len(Xtop)-1):                
                self.add(TopChord([Xtop[i],Xtop[i+1]],copy(top_chord_profile),nel=nel_chord))
        
            # Generate bottom chord members
            for i in range(len(Xbot)-1):
                self.add(BottomChord([Xbot[i],Xbot[i+1]],copy(bottom_chord_profile),nel=nel_chord))
            
            # Generate braces
            if dx1 == 0:
                if first_diagonal_up:
                    # Verticals
                    if edge_verticals:
                        for X, Y in zip(Xtop[::2],Xbot):
                            self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                    else:
                        for X, Y in zip(Xtop[2::2],Xbot[1:-1]):
                            self.add(TrussBrace([X,Y],copy(brace_profile),nel=nel_brace))
                
                    # Diagonals
                    self.add(TrussBrace([Xbot[0],Xtop[1]],copy(brace_profile),nel=nel_brace))                    
                    
                    for i in range(1,len(Xbot)-1):
                        self.add(TrussBrace([Xtop[2*i-1],Xbot[i]],copy(brace_profile),nel=nel_brace))
                        self.add(TrussBrace([Xtop[2*i+1],Xbot[i]],copy(brace_profile),nel=nel_brace))
                        
                    
                    self.add(TrussBrace([Xbot[-1],Xtop[-2]],copy(brace_profile),nel=nel_brace))
                else:   
                    if edge_verticals:
                        self.add(TrussBrace([Xbot[0],Xtop[0]],copy(brace_profile),nel=nel_brace))
                        self.add(TrussBrace([Xbot[-1],Xtop[-1]],copy(brace_profile),nel=nel_brace))
                   
                    for i in range(1,len(Xbot)-1):                           
                        self.add(TrussBrace([Xtop[2*i-2],Xbot[i]],copy(brace_profile),nel=nel_brace))
                        self.add(TrussBrace([Xtop[2*i-1],Xbot[i]],copy(brace_profile),nel=nel_brace))
                        self.add(TrussBrace([Xtop[2*i],Xbot[i]],copy(brace_profile),nel=nel_brace))
                                       
                    
                   
                    """
                    if first_diagonal_up:
                        for i in range(1,len(Xbot)-2):                           
                            self.add(TrussBrace([Xtop[2*i-2],Xbot[i]],brace_profile))
                            self.add(TrussBrace([Xtop[2*i-1],Xbot[i]],brace_profile))
                            self.add(TrussBrace([Xtop[2*i+1],Xbot[i]],brace_profile))
                    else:
                        for i in range(1,len(Xbot)-1):                           
                            self.add(TrussBrace([Xtop[2*i-2],Xbot[i]],brace_profile))
                            self.add(TrussBrace([Xtop[2*i-1],Xbot[i]],brace_profile))
                            self.add(TrussBrace([Xtop[2*i],Xbot[i]],brace_profile))
                       
                        self.add(TrussBrace([Xbot[-1],Xtop[-1]],brace_profile))
                    """
            else:
                # Lower chord is shorter than top chord. Now the first
                # diagonal is downward anyway.
                
                # Verticals
                #for X, Y in zip(Xtop[1::2],Xbot):
                #    self.add(TrussBrace([X,Y],brace_profile))
                    
                for i in range(len(Xbot)):
                    #print(i)                       
                    self.add(TrussBrace([Xtop[2*i],Xbot[i]],copy(brace_profile),nel=nel_brace))
                    self.add(TrussBrace([Xtop[2*i+1],Xbot[i]],copy(brace_profile),nel=nel_brace))
                    self.add(TrussBrace([Xtop[2*i+2],Xbot[i]],copy(brace_profile),nel=nel_brace))
        """    
        fig, ax = plt.subplots(1)
        
        ax.plot([X[0] for X in Xtop], [X[1] for X in Xtop],'ob')
        ax.plot([X[0] for X in Xbot], [X[1] for X in Xbot],'ob')
        ax.set_aspect('equal')
        """  
        self.top_nodes = Xtop
        self.bottom_nodes = Xbot
    def generate_supports(self):
        """ Creates supports to the truss """
        
        if self._dx_left > 0:
            # If the first node of the bottom chord is not at the left edge,
            # the first node of the top chord is supported
            self.add(XYHingedSupport(self.top_nodes[0]))
        else:
            # Support the first node of the bottom chords
            self.add(XYHingedSupport(self.bottom_nodes[0]))
        
        if self._dx_right > 0:
            self.add(YHingedSupport(self.top_nodes[-1]))
        else:
            self.add(YHingedSupport(self.bottom_nodes[-1]))
    
    def generate_uniform_load(self,q,chord="top",load_id=LoadIDs['ULS'],ltype='live'):
        """ Creates a uniform line load to one of the chords """

        vals = [q,q]

        if chord == "top":
            load_name = 'TC_' + ltype + '_line'
            for mem in self.top_chord:
                self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name, "global"))
        elif chord == "bottom":          
            load_name = 'BC_' + ltype + '_line'
            for mem in self.bottom_chord:
                self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name, "global"))
    
    def generate_piecewise_uniform_load(self,q,xzones,load_id=LoadIDs['ULS'],ltype='live'):
        """ Creates piecewise uniform line load which affects the top chord 
            :param q: list of load values
            :param xzones: list of x-coordinates of the wind zones
        """
        load_name0 = 'TC_' + ltype
        
        for mem in self.top_chord:
            if mem.nodes[1].x <= xzones[0]:
                # Member is in the first zone
                vals = [q[0],q[0]]
                load_name = load_name0 + '_line'
                self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name, "global"))
            elif mem.nodes[0].x > xzones[-1]:
                # Member is in the last zone
                vals = [q[-1],q[-1]]
                load_name = load_name0 + '_line'
                self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name, "global"))
            else:
                
                for i, (x1, x2) in enumerate(zip(xzones[:-1],xzones[1:])):
                    if mem.nodes[0].x < x1 and mem.nodes[1].x > x1:
                        # member is partially outside
                        vals = [q[i],q[i+1]]
                        xdiv = (x1-mem.nodes[0].x)/(mem.nodes[1].x-mem.nodes[0].x)
                        #xdiv = x1
                        load_name = load_name0 + '_pw_line'
                        self.add(PWLineLoad(mem, vals, xdiv, 'y', load_id, 1.0, ltype, load_name,"global"))
                    elif mem.nodes[0].x >= x1 and mem.nodes[1].x > x2:
                        # member is partially outside
                        vals = [q[i+1],q[i+2]]
                        xdiv = (x2-mem.nodes[0].x)/(mem.nodes[1].x-mem.nodes[0].x)
                        #xdiv = x2
                        load_name = load_name0 + '_pw_line'
                        self.add(PWLineLoad(mem, vals, xdiv, 'y', load_id, 1.0, ltype, load_name,"global"))
                    elif mem.nodes[0].x >= x1 and mem.nodes[1].x <= x2:
                        # member is totally inside
                        vals = [q[i+1],q[i+1]]
                        load_name = load_name0 + '_line'
                        self.add(LineLoad(mem, vals, 'y', load_id, 1.0, ltype, load_name,"global"))
    
    def generate_joints(self):
        """ Creates joints for the truss. These are welded tubular joints by default. """
        
        joints = {}
        
        # First, determine which members are connected to each node
        # This information is stored in 'joints' dict.
        for node in self.nodes:
            new_joint = {'chord':[],'braces':[]}
            for mem in self.top_chord:
                if node in mem.nodes:
                    new_joint['chord'].append(mem)
                    if len(new_joint['chord'])==2:
                        break                    
            
            for mem in self.bottom_chord:
                if node in mem.nodes:
                    new_joint['chord'].append(mem)
                    if len(new_joint['chord'])==2:
                        break
            
            for mem in self.braces.values():
                if node in mem.nodes:
                    new_joint['braces'].append(mem)
                    # The brace knows to which joints it belongs.
                    if node in self.top_nodes:
                        mem.top_joint = new_joint
                    else:
                        mem.bottom_joint = new_joint
            
            
            joints[node] = new_joint
        
        # Create joints. As a first guess, the joint type is
        # based on number of braces. This may not be accurate for
        # determining X joints, because both X and K joints have two braces.
        # The user must probably provide some additional information here to
        # distinguish the proper joint type, or the type will be based on
        # the results of structural analysis.
        njoint = 0
        for node, joint in joints.items():
            
            nbraces = len(joint['braces'])
            
            if nbraces == 3:
                newJoint = TubularKTGapJoint(node, joint['chord'], joint['braces'])                 
                self.add(newJoint)                
                njoint += 1
                
                if node in self.top_nodes and node.x == self.L1:
                    self.ridge_joint = newJoint
                
            elif nbraces == 2:
                # K joint
                newJoint = TubularKGapJoint(node, joint['chord'], joint['braces'])                 
                self.add(newJoint)                
                njoint += 1
                
                if node in self.top_nodes and node.x == self.L1:
                    self.ridge_joint = newJoint
                    
            elif nbraces == 1:                
                newJoint = TubularYJoint(node, joint['chord'], joint['braces'])
                self.add(newJoint)
                njoint += 1
            elif nbraces == 0:
                # Number of braces can be 0, if top or bottom chord is supported
                # at the node
                newJoint = None
            
            # Attach the joint to the corresponding braces
            for brace in joint['braces']:
                if node in self.top_nodes:
                    brace.top_joint = newJoint
                else:
                    brace.bottom_joint = newJoint
            
            # Attach the joint to the corresponding chord members
            if not newJoint is None:
                for chord in joint['chord']:                
                    chord.joints.append(newJoint)
        
        # Set the actual geometry for all joints
        for joint in self.joints.values():                      
            if isinstance(joint,TubularYJoint):
                angle = joint.brace_angle(joint.braces[0],geometry="actual")
                joint.joint.angle = angle
            elif isinstance(joint,TubularKGapJoint):
                # The gap calculation routine updates the
                # brace angles and gap
                #print(joint.gap)
                joint.calc_gap()                
                #print(joint.gap)
    
    def symmetry(self,axis='y',sym_value=None):
        
        # Make general actions for symmetry
        super().symmetry(axis,sym_value)
        
        # Symmetry of joints
        #joint_list = list(self.joints.values())
        for joint in self.joints.values():
            #print(f"Symmetry for joint {joint.joint_id}.")
            sym_node = joint.symmetry_joint(self.joints)
            #print(sym_node)
    
    def clear_fem(self):
        """ Clears FEM data """
        super().clear_fem()
        
        for joint in self.joints.values():
            joint.clear_fem()
        
    def generate_fem(self,model="no_eccentricity",nodes_and_elements_only=False):
        """
        Creates FEM model

        Parameters
        ----------
        model : string, optional
            Type of FEM model. Allowed values are:
                "simple": all members are pin-jointed bars
                "no_eccentricity": braces meet at originally created nodes
                "en1993": for K joints, the braces meet at the intersection point of their
                        centerlines, and an eccentricity element is added
                "braces2chord_faces": braces end at chord faces.
                The default is "no_eccentricity".

        Returns
        -------
        None.

        """
                
        
        if model == "no_eccentricity":
            # Here, we can use the fem model generation of Raami class
            super().generate_fem()
            
            for joint in self.joints.values():
                # Generate fem nodes for the joints
                x = joint.node.coords
                for node in self.fem.nodes:
                    if np.linalg.norm(node.coord-x) < 1e-6:
                        joint.fem_nodes["xc"] = node
                        break
                            
        elif model == "ecc_elements":
            # Eccentricity elements are added for K joints
            # from the intersection of brace centerline and chord
            # face to the chord centerline.
            for joint in self.joints.values():
                if isinstance(joint,TubularYJoint):      
                    # For Y joints, no eccentricity elements are used.
                    newNode = self.fem.add_node(joint.node.x,joint.node.y)                    
                    joint.fem_nodes['xc'] = newNode
                    joint.node.fem_node = newNode
                elif isinstance(joint,TubularKGapJoint):
                    # For K gap joints, we introduce the following FEM nodes:
                    # 1) intersection of each brace and chord face
                    # 2) chord centerline points closest to the chord face points.
                    
                    
                    """
                    if joint is self.ridge_joint:
                        xc = joint.node.coords
                        xcNode = self.fem.add_node(xc[0],xc[1])
                        joint.fem_nodes['xc'] = xcNode
                        
                        xface, xcenter = joint.brace_chord_face_x()
                        print(xface,xcenter)
                    else:
                        xface, xcenter = joint.brace_chord_face_x()
                        
                        for brace, xf in xface.items():
                            newNode = self.fem.add_node(xf[0],xf[1])
                            joint.fem_nodes['xcf'][brace] = newNode
                        
                        for brace, xc in xcenter.items():
                            newNode = self.fem.add_node(xc[0],xc[1])
                            joint.fem_nodes['xch'][brace] = newNode
                    """
                    xface, xcenter = joint.brace_chord_face_x()
                    for brace, xf in xface.items():
                        newNode = self.fem.add_node(xf[0],xf[1])
                        joint.fem_nodes['xcf'][brace] = newNode
                    
                    for brace, xc in xcenter.items():
                        newNode = self.fem.add_node(xc[0],xc[1])
                        joint.fem_nodes['xch'][brace] = newNode
                    
                    # For the apex, create a FEM node to the intersection
                    # of chord centerlines
                    if joint is self.ridge_joint:
                        xc = joint.node.coords
                        xcNode = self.fem.add_node(xc[0],xc[1])
                        joint.fem_nodes['xc'] = xcNode
                        joint.node.fem_node = xcNode
                    
                    """
                    xpNode = self.fem.add_node(xp[0],xp[1])
                    if joint is self.ridge_joint:
                        xc = joint.node.coords
                    else:
                        xc = joint.xchord()
                    """
                    
                    """
                    if np.linalg.norm(xp-xc) < 1e-6:
                        xcNode = xp
                    else:
                        xcNode = self.fem.add_node(xc[0],xc[1])
                        
                    joint.fem_nodes['xc'] = xcNode
                    joint.fem_nodes['xp'] = xpNode
                    joint.node.fem_node = newNode
                    """
                elif isinstance(joint,TubularKTGapJoint):
                    # For KT gap joints, currently only a node in the
                    # initial FrameNode location is created.
                    newNode = self.fem.add_node(joint.node.x,joint.node.y)                    
                    joint.fem_nodes['xc'] = newNode
                    joint.node.fem_node = newNode
                                    
            for mem in self.braces.values():
                #Generate member elements                
                mem.generate_elements(self.fem,model,self.braces_as_beams)
                                    
            for mem in self.top_chord:
                mem.generate_elements(self.fem,model)
        
            for mem in self.bottom_chord:
                mem.generate_elements(self.fem,model)
            
            # TODO! Change eccentricity elements to rigid links!
            # (Is it needed, or do IPE 500 elements suffice?)
            for joint in self.joints.values():
                if isinstance(joint,TubularKGapJoint):
                    # Eccentricity elements
                    n1 = joint.fem_nodes['xcf'][joint.left_brace]
                    nc1 = joint.fem_nodes['xch'][joint.left_chord]
                    cs = IPE(500)
                                 
                    newElement = EBBeam(n1,nc1,cs,cs.material)                    
                    
                    self.fem.add_element(newElement)
                    joint.fem_elements['left_ecc'] = newElement
                    
                    n1 = joint.fem_nodes['xcf'][joint.right_brace]
                    nc2 = joint.fem_nodes['xch'][joint.right_chord]
                                 
                    newElement = EBBeam(n1,nc2,cs,cs.material)
                    
                    joint.fem_elements['right_ecc'] = newElement
                    
                    self.fem.add_element(newElement)
                    
                    # Add gap elements
                    if joint.ridge:
                        nApex = joint.fem_nodes['xc']
                        cs_chord = joint.chords[0].cross_section
                        newElement1 = EBBeam(nc1,nApex,cs_chord,cs_chord.material)
                        self.fem.add_element(newElement1)
                        newElement2 = EBBeam(nApex,nc2,cs_chord,cs_chord.material)
                        self.fem.add_element(newElement2)
                        joint.fem_elements['gap'] = [newElement1,newElement2]
                    else:
                        cs_chord = joint.chords[0].cross_section
                        newElement = EBBeam(nc1,nc2,cs_chord,cs_chord.material)
                        self.fem.add_element(newElement)
                        joint.fem_elements['gap'] = newElement
            
            if not nodes_and_elements_only:
                # Generate supports
                for supp in self.supports.values():
                    supp.add_support(self.fem)
                                
                # Generate loads
                for load in self.loads.values():
                    # add_load addes the loads on the target member
                    load.add_load(self.fem)
                    # Creates new loadcase if one with same load_id does not exist
                    lcase_ids = [lc.load for lc in self.fem.loadcases.values()]
                    if load.load_id not in lcase_ids:
                        self.fem.add_loadcase(supp_id=1,load_id=load.load_id)
                
                # Add loads to gap elements
                for joint in self.joints.values():
                    if isinstance(joint,TubularKGapJoint):
                        # Add load to gap element
                        for load1 in joint.chords[0].loads:
                            for load2 in joint.chords[1].loads:
                                if load1.load_id == load2.load_id:
                                    if joint == self.ridge_joint:
                                        q = load1.f*load1.values[1]
                                        load = FEMLineLoad(load1.load_id,
                                                           joint.fem_elements['gap'][0],
                                                           [0.0,1.0],
                                                           [q,q],
                                                           load1.direction,
                                                           load1.coord_sys)
                                        self.fem.add_load(load)
                                        q = load1.f*load2.values[0]
                                        load = FEMLineLoad(load1.load_id,
                                                           joint.fem_elements['gap'][1],
                                                           [0.0,1.0],
                                                           [q,q],
                                                           load1.direction,
                                                           load1.coord_sys)
                                        self.fem.add_load(load)
                                    else:
                                        if isinstance(load1,LineLoad) or isinstance(load1,PWLineLoad):
                                            if abs(load1.values[1]) >= abs(load2.values[0]):
                                                q = load1.f*load1.values[1]
                                                
                                            else:
                                                q = load2.f*load2.values[0]
                                                                                    
                                            load = FEMLineLoad(load1.load_id,
                                                               joint.fem_elements['gap'],
                                                               [0.0,1.0],
                                                               [q,q],
                                                               load1.direction,
                                                               load1.coord_sys)                                            
                                            self.fem.add_load(load)
                                    break
                                    
                
                # Set nodal degrees of freedom
                self.fem.nodal_dofs()
                
        else:
            print("Weird FEM model.")
    
    def update_fem_joint_nodes(self):
        """ Updates nodal positions at joints. It is assumed that cross-sections
            have been changed along the way.
        """
        
        for joint in self.joints.values():
            if isinstance(joint,TubularKGapJoint):
                # Calculate brace-to-chord face intersection points and the
                # corresponding projections to the centerline of the chord
                xface, xcenter = joint.brace_chord_face_x()
                for brace, xf in xface.items():                    
                    joint.fem_nodes['xcf'][brace].x = xf[0]
                    joint.fem_nodes['xcf'][brace].y = xf[1]
                
                for brace, xc in xcenter.items():                    
                    joint.fem_nodes['xch'][brace].x = xc[0]
                    joint.fem_nodes['xch'][brace].y = xc[1]
    
    def plot(self, geometry=False, print_text=True, show=True,
             loads=True, color=False, axes=None, save=False, mem_dim=False,
             saveopts={'filename':'default','format':'svg','orientation':'landscape', 'papertype':'a3'}):
        """ Plots the frame
            
            Parameters
            ----------
            :param print_text: Set true to print member's profiles and names (default: True)
            :param show: Set true to show the plot (default: True)
            :param loads: Set true to show loads (default: True)
            :param color: Set true to show members' utilization ratio (default: False)

            :type print_text : bool
            :type show: bool
            :type loads: bool
            :type color: bool

            Colors' meaning:

                blue -- member has load
                green -- member can bear its loads
                red -- member breaks under its loads
                black -- member is added, but not designed
        """
        if not geometry:
            super().plot(print_text, show, loads, color, axes, save, mem_dim)
        else:
            
            if axes is None:
                fig, ax = plt.subplots(1)
            else:
                ax = axes
    
            # Plot members
            plt.rcParams['text.color'] = "red"
            plt.rcParams['font.size'] = 2.5
            for mem in self.top_chord:
                mem.plot(print_text, color, ax, mem_dim=True)
            
            for mem in self.bottom_chord:
                mem.plot(print_text, color, ax, mem_dim=True)
            
            for mem in self.braces.values():                            
                mem.plot(print_text, color, ax, mem_dim=True, geometry=geometry)
    
            # Plot joints
            #for joint in self.joints.values():
            #    joint.plot(color=color)
    
            # Plot supports
            
            for support in self.supports.values():
                node_coord = support.node.coords
                if support.dofs == [-1, -1, -1] or support.dofs == [1, 1, 1] or support.dofs == [0, 1, 2]:
                    marker = 's'
                elif support.dofs == [1]:
                    marker = '^'
                elif support.dofs == [0]:
                    marker = '>'
                else:
                    marker = 'D'
                ax.scatter(node_coord[0], node_coord[1], s=50, c='k',
                            marker=marker)
                
            ax.axis('equal')
            if save:
                #plt.savefig('default.svg', format='svg')
                plt.savefig(saveopts['filename'],format=saveopts['format'],\
                            orientation=saveopts['orientation'],papertype=saveopts['papertype'])
            if show:
                plt.axis('equal')
                plt.show()
                
    def optimize_members(self, prof_type="CURRENT",verb=False,**kwargs):    
        """ Finds minimum weight profiles for members """
               
        top = {'material': 'S355', 'class': 2, 'utility': 1.0, 'bmin':10, 'bmax':1e5}
        bottom = {'material': 'S355', 'class': 2, 'utility': 1.0,'bmin':10, 'bmax':1e5}
        braces = {'material': 'S355', 'class': 2, 'utility_tens': 1.0, 'utility_comp':1.0, 'bmin':10, 'bmax':1e5}
            
        limit_width = False
        
        for key, value in kwargs.items():
            if key == 'top':
                top = value
            elif key == 'bottom':
                bottom = value
            elif key == 'braces':
                braces = value
            elif key == 'limit_width':
                limit_width = value
        
        kmax = 10
        k = 1
                        
        while k < kmax:            
            explored = []
            MEM_PROFILES_CHANGED = []
            self.structural_analysis('all','REM')
            if verb:
                print(f"Optimize profiles, iteration {k}")
            
            if limit_width:
                b_chord_min = self.max_brace_width()
                #b_brace_max = min(self.top_chord_profile.b,self.bottom_chord_profile.b)
                b_brace_max = 1e5
            else:
                b_chord_min = 10
                b_brace_max = 1e5
            
            # Go through groups first
            for name, group in self.member_groups.items():
                if name == "top_chord":
                    material = top['material']
                    sec_class = top['class']
                    max_utility = top['utility']                    
                elif name == 'bottom_chord':
                    material = bottom['material']
                    sec_class = bottom['class']
                    max_utility = bottom['utility']                    
                    
                group.optimum_design(prof_type,verb,material,sec_class,max_utility,b_chord_min)
                
                for mem in group.members:
                    explored.append(mem)
                
            
            for member in self.members.values():
                #print(member)                
                if not member in explored:
                    if verb:
                        print(member)
                    explored.append(member)
                    if isinstance(member,TrussBrace):
                        material = braces['material']
                        sec_class = braces['class']
                        if list(member.NEd.values())[0] >= 0.0:
                            max_utility = braces['utility_tens']
                        else:
                            max_utility = braces['utility_comp']
                    else:
                        material = 'S355'
                        sec_class = 2
                        max_utility = 1.0
                        
                    MEM_PROFILES_CHANGED.append(member.optimum_design(prof_type,verb,material,sec_class,max_utility,bmax=b_brace_max))
                                        
                    if not member.symmetry_pair is None:
                        explored.append(member.symmetry_pair)
                    
                    #print(member.cross_section)
            
            # If none of the members changed profile, exit the loop
            if not any(MEM_PROFILES_CHANGED):
                break
                
            k += 1
        
        if verb:
            print(f"Number of iterations: {k}.")
        
        # First run the optimization method of Raami class
        # super().optimize_members(prof_type,verb)
        
def vierendeel_example(L=20000,h=2000,panels=8):
        
    t = VierendeelTruss(span=L, height=h, panels=panels)
    t.generate_topology(nel_chord=6,nel_brace=6)
    t.generate_supports()
    t.generate_joints()
    #t.generate_uniform_load(q=-25)
    P = -17e3
    t.generate_chord_point_load(p=[0.5*P,P])
    t.generate_fem(model='no_eccentricity')
    t.structural_analysis(load_id=t.load_ids[0],support_method="REM")
    t.bmd(scale=1,load_id=t.load_ids[0],loads=False)
    t.optimize_members(verb=True,top={'material':'S700','class':2,'utility':1.0},
                       bottom={'material':'S700','class':2,'utility':1.0},
                       braces = {'material': 'S700', 'class': 2, 'utility_tens': 1.0, 'utility_comp':1.0},
                       limit_width=True)
    t.bmd(scale=1,load_id=t.load_ids[0],loads=False)
    return t
    #t.plot(mem_dim=True)


if __name__ == "__main__":
    
    
    from timeit import default_timer as timer
    
    t = vierendeel_example(L=6*3000,h=2500,panels=6)

    t.plot(loads=False)
    
    """
    t.symmetry()
    
   
    t.generate_supports()
    t.generate_joints()
    #print(t.joints[1].gap)
    #print(t.top_nodes)
    t.generate_uniform_load(q=-25)
    t.generate_fem(model='en1993')
    #t.generate_fem(model='no_eccentricity')
    t.structural_analysis(load_id=t.load_ids[0],support_method="REM")
    
    #t.bmd(scale=20,load_id=t.load_ids[0],loads=False)
    #t.cost()
    #start = timer()
    #t.optimize_members(verb=True,top={'material':'S700','class':2},
    #                   bottom={'material':'S700','class':2},
    #                   braces={'material':'S700','class':2})
    #end = timer()
    #print(f"Time elapsed: {end-start:.2f} s")
    
    t.structural_analysis(load_id=t.load_ids[0],support_method="REM")
    
    t.plot(geometry=True)
    #t.plot(mem_dim=True)
    """
    #opts = {'x_monitor':0.5*t.span, 'n_monitored':2}
    #t.to_abaqus(filename='KT-ristikko',partname="KT-ristikko",options=opts)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  3 21:22:03 2020

@author: kmela
"""

import math
import numpy as np
from metku.structures.steel.steel_member import SteelMember

class CFSteelMember(SteelMember):
    """ Class for cold-formed steel members """
    
    def __init__(self, profile, length, Lcr=[1.0, 1.0], mtype="beam",
                 LT_buckling=False):
        
        super().__init__(profile,length,Lcr,mtype,LT_buckling)
    
    def slenderness(self, verb=False):
        """ Non-dimensional slenderness according to EN 1993-1-1 """
        NRd = self.profile.Aeff * self.fy()
        Ncr = self.ncrit(verb)
        # if NRd <= 1e-6:
        # print(self.profile.A)
        # print(self.profile.h)
        # print(NRd, Ncr)

        slend = np.sqrt(NRd / Ncr)

        if verb:
            print("lambda,y = {0:4.2f}".format(slend[0]))
            print("lambda,z = {0:4.2f}".format(slend[1]))

        return slend
    
    def buckling_strength(self, verb=False):
        """ Member buckling strength according to EN 1993-1-3 """

        if verb:
            print("** Buckling strength** ")

        slend = self.slenderness(verb)
        NbRd = []
        NRd = self.profile.Aeff * self.fy()
        # self.profile.imp_factor()
        # print(slend)
        # p = 0.5*(1+np.array(self.profile.imp_factor)*(slend-0.2)+slend**2)
        # r = 1/(p+np.sqrt(p**2-slend**2))


        for i in range(len(slend)):
            p = 0.5 * (1 + self.profile.imp_factor[i] * (slend[i] - 0.2) +
                       slend[i] ** 2)
            r = min(1 / (p + math.sqrt(p ** 2 - slend[i] ** 2)), 1.0)
            if verb:
                print("Psi,{0:1} = {1:4.2f}".format(i, p))
                print("chi,{0:1} = {1:4.2f}".format(i, r))
                print("NbRd,{0:1} = {1:4.2f}".format(i, NRd * r * 1e-3))

            NbRd.append(NRd * r)

        # NbRd = self.profile.A*self.fy()*r;
        return NbRd    
    
    def Sreq(self):
        """ Required shear stiffness for sheeting to provide full
            lateral support to the purlin
        """
        
        Sw = self.profile.E*self.profile.Iw*math.pi**2/self.length**2
        St = self.profile.G*self.profile.It
        Sz = self.profile.E*self.profile.Iz*math.pi**2/self.length**2*0.25*self.profile.h**2
        
        
        
        return (Sw + St + Sz)*70/self.profile.h**2
    
    
    
if __name__ == '__main__':
    
    from eurocodes.en1993.en1993_1_3 import cf_profs
    from eurocodes.en1993.en1993_1_3.en1993_1_3 import sheeting_shear_stiffness
    
    c = cf_profs.CSection(t_nom=2.0,h=120,a=60,b=60,ca=20,cb=20,r=2.0)
    z = cf_profs.ruukki_z(300,2.5)
    m = CFSteelMember(z,length=5000)
    
    #print(m.kh())
    
    #print(c.kh())
    #print(m.Sreq())
    #print(sheeting_shear_stiffness(t=0.96,s=1500,hw=(70-1),broof=24000))